import unittest

from rpft.rapidpro.models.containers import RapidProContainer, FlowContainer
from rpft.rapidpro.models.actions import Group, AddContactGroupAction
from rpft.rapidpro.models.nodes import BasicNode, SwitchRouterNode, EnterFlowNode
from rpft.rapidpro.models.campaigns import Campaign, CampaignEvent
from rpft.rapidpro.models.triggers import Trigger


def get_flow_with_group_and_flow_node():
    flow = FlowContainer("Group Flow")
    node = BasicNode()
    node.add_action(
        AddContactGroupAction(
            [Group("No UUID Group"), Group("UUID Group", "fake-uuid")]
        )
    )
    flow.add_node(node)
    flow.add_node(EnterFlowNode("Second Flow"))
    return flow


def get_has_group_flow():
    flow = FlowContainer("Second Flow")
    node = SwitchRouterNode("@contact.groups")
    node.add_choice(
        comparison_variable="@contact.groups",
        comparison_type="has_group",
        comparison_arguments=[None, "No UUID Group"],
        category_name="No Category Name",
        destination_uuid=None,
    )
    node.add_choice(
        comparison_variable="@contact.groups",
        comparison_type="has_group",
        comparison_arguments=[None, "UUID Group"],
        category_name="Category Name",
        destination_uuid=None,
    )
    flow.add_node(node)
    flow.add_node(EnterFlowNode("Second Flow"))
    return flow


class TestRapidProContainer(unittest.TestCase):
    def setUp(self) -> None:
        pass

    def test_assign_group_and_flow(self):
        rpc = RapidProContainer()
        rpc.add_flow(get_flow_with_group_and_flow_node())
        self.assertIsNone(rpc.flows[0].nodes[0].actions[0].groups[0].uuid)
        self.assertEqual(rpc.flows[0].nodes[0].actions[0].groups[1].uuid, "fake-uuid")
        self.assertIsNone(rpc.flows[0].nodes[1].actions[0].flow.uuid)
        rpc.update_global_uuids()
        self.assertIsNotNone(rpc.flows[0].nodes[0].actions[0].groups[0].uuid)
        self.assertEqual(rpc.flows[0].nodes[0].actions[0].groups[1].uuid, "fake-uuid")
        self.assertIsNotNone(rpc.flows[0].nodes[1].actions[0].flow.uuid)

    def test_assign_group_predefined(self):
        rpc = RapidProContainer(groups=[Group("No UUID Group", "ABCD")])
        rpc.add_flow(get_flow_with_group_and_flow_node())
        self.assertIsNone(rpc.flows[0].nodes[0].actions[0].groups[0].uuid)
        rpc.update_global_uuids()
        self.assertEqual(rpc.flows[0].nodes[0].actions[0].groups[0].uuid, "ABCD")

    def test_assign_group_clash(self):
        rpc = RapidProContainer(groups=[Group("UUID Group", "ABCD")])
        rpc.add_flow(get_flow_with_group_and_flow_node())
        self.assertEqual(rpc.flows[0].nodes[0].actions[0].groups[1].uuid, "fake-uuid")
        with self.assertRaises(ValueError):
            # ValueError: Group/Flow UUID Group has multiple uuids: fake-uuid and ABCD
            rpc.update_global_uuids()

    def test_consistency(self):
        rpc = RapidProContainer()
        rpc.add_flow(get_flow_with_group_and_flow_node())
        rpc.add_flow(get_has_group_flow())
        rpc.update_global_uuids()
        self.assertEqual(rpc.flows[1].uuid, rpc.flows[0].nodes[1].actions[0].flow.uuid)
        self.assertEqual(
            rpc.flows[1].nodes[0].router.cases[0].arguments[0],
            rpc.flows[0].nodes[0].actions[0].groups[0].uuid,
        )
        self.assertEqual(
            rpc.flows[1].nodes[0].router.cases[1].arguments[0], "fake-uuid"
        )

    def test_assign_uuids_to_campaign(self):
        rpc = RapidProContainer()
        rpc.add_flow(get_flow_with_group_and_flow_node())
        event = CampaignEvent(
            offset=1234,
            unit="H",
            event_type="F",
            delivery_hour=-1,
            start_mode="I",
            relative_to_label="Created On",
            flow_name="Second Flow",
        )
        campaign = Campaign("My Campaign", Group("UUID Group"), events=[event])
        rpc.add_campaign(campaign)
        rpc.update_global_uuids()
        self.assertEqual(rpc.campaigns[0].group.uuid, "fake-uuid")
        self.assertEqual(
            rpc.campaigns[0].events[0].flow.uuid,
            rpc.flows[0].nodes[1].actions[0].flow.uuid,
        )

    def test_get_uuids_from_campaign(self):
        rpc = RapidProContainer()
        rpc.add_flow(get_flow_with_group_and_flow_node())
        event = CampaignEvent(
            offset=1234,
            unit="H",
            event_type="F",
            delivery_hour=-1,
            start_mode="I",
            relative_to_label="Created On",
            flow_name="Second Flow",
            flow_uuid="fake-flow-uuid",
        )
        campaign = Campaign(
            "My Campaign", Group("No UUID Group", "fake-group-uuid"), events=[event]
        )
        rpc.add_campaign(campaign)
        rpc.update_global_uuids()
        self.assertEqual(rpc.campaigns[0].group.uuid, "fake-group-uuid")
        self.assertEqual(
            rpc.flows[0].nodes[0].actions[0].groups[0].uuid, "fake-group-uuid"
        )
        self.assertEqual(rpc.flows[0].nodes[1].actions[0].flow.uuid, "fake-flow-uuid")
        self.assertEqual(rpc.campaigns[0].events[0].flow.uuid, "fake-flow-uuid")

    def test_get_uuids_from_trigger(self):
        rpc = RapidProContainer()
        rpc.add_flow(get_flow_with_group_and_flow_node())
        trigger = Trigger(
            "K",
            "keyword",
            flow_name="Second Flow",
            flow_uuid="fake-flow-uuid",
            group_names=["No UUID Group"],
            group_uuids=["fake-group-uuid"],
        )
        rpc.add_trigger(trigger)
        rpc.update_global_uuids()
        self.assertEqual(rpc.triggers[0].groups[0].uuid, "fake-group-uuid")
        self.assertEqual(
            rpc.flows[0].nodes[0].actions[0].groups[0].uuid, "fake-group-uuid"
        )
        self.assertEqual(rpc.flows[0].nodes[1].actions[0].flow.uuid, "fake-flow-uuid")
        self.assertEqual(rpc.triggers[0].flow.uuid, "fake-flow-uuid")



import unittest
from rpft.parsers.creation.tagmatcher import TagMatcher


class TestTagMatcher(unittest.TestCase):
    def test_empty_matcher(self):
        tm = TagMatcher()
        self.assertTrue(tm.matches([]))
        self.assertTrue(tm.matches([""]))
        self.assertTrue(tm.matches(["", ""]))
        self.assertTrue(tm.matches(["tag1", "tag2"]))

    def test_single_tag_position_inclusion(self):
        tm = TagMatcher(["1", "foo", "bar"])
        self.assertTrue(tm.matches([]))
        self.assertTrue(tm.matches([""]))
        self.assertTrue(tm.matches(["", ""]))
        self.assertTrue(tm.matches(["", "something"]))
        self.assertTrue(tm.matches(["foo"]))
        self.assertTrue(tm.matches(["bar"]))
        self.assertTrue(tm.matches(["bar", "something"]))
        self.assertFalse(tm.matches(["something"]))
        self.assertFalse(tm.matches(["something", "foo"]))
        self.assertFalse(tm.matches(["something", "bar"]))

    def test_multiple_tag_position_inclusion(self):
        tm = TagMatcher(["1", "foo", "bar", "2", "baz"])
        self.assertTrue(tm.matches([]))
        self.assertTrue(tm.matches([""]))
        self.assertTrue(tm.matches(["", ""]))
        self.assertTrue(tm.matches(["", "baz"]))
        self.assertTrue(tm.matches(["foo", ""]))
        self.assertTrue(tm.matches(["foo", "baz"]))
        self.assertTrue(tm.matches(["bar", "baz"]))
        self.assertTrue(tm.matches(["foo"]))
        self.assertTrue(tm.matches(["bar"]))
        self.assertFalse(tm.matches(["bar", "something"]))
        self.assertFalse(tm.matches(["something"]))
        self.assertFalse(tm.matches(["something", "foo"]))
        self.assertFalse(tm.matches(["something", "baz"]))

    def test_tag_exclusion(self):
        tm = TagMatcher(["1", "!foo", "!bar"])
        self.assertTrue(tm.matches([]))
        self.assertTrue(tm.matches([""]))
        self.assertTrue(tm.matches(["baz"]))
        self.assertFalse(tm.matches(["bar"]))
        self.assertFalse(tm.matches(["foo"]))

    def test_exclude_empty_only(self):
        tm = TagMatcher(["1", "!"])
        self.assertTrue(tm.matches([]))
        self.assertTrue(tm.matches(["bar"]))
        self.assertFalse(tm.matches([""]))

    def test_tag_inclusion_exclude_empty(self):
        tm = TagMatcher(["1", "!", "foo", "bar"])
        self.assertTrue(tm.matches([]))
        self.assertTrue(tm.matches(["bar"]))
        self.assertTrue(tm.matches(["foo"]))
        self.assertFalse(tm.matches(["baz"]))
        self.assertFalse(tm.matches([""]))



from rpft.parsers.creation.datarowmodel import DataRowModel
from rpft.parsers.creation.models import SurveyQuestionModel


class SurveyQuestionRowModel(DataRowModel, SurveyQuestionModel):
    pass


class IDValueRowModel(DataRowModel):
    value: str = ""



from unittest import TestCase

from rpft.parsers.sheets import DatasetSheetReader
from rpft.parsers.universal import (
    bookify,
    parse_cell,
    parse_table,
    parse_tables,
    stringify,
    tabulate,
)
from tablib import Dataset


class TestConvertDataToCell(TestCase):
    def test_delimiters_can_be_configured(self):
        self.assertEqual(
            stringify(
                [
                    ["click", ["auth", "sign_in_google"]],
                    ["click", ["emit", "force_reprocess"]],
                ],
                delimiters=";|:",
            ),
            "click | auth : sign_in_google ; click | emit : force_reprocess",
        )


class TestConvertUniversalToTable(TestCase):
    def test_headers_must_be_first_row(self):
        data = [
            {"type": "create_flow", "sheet_name": "flow1"},
        ]

        table = tabulate(data)

        self.assertEqual(
            table[0],
            ["type", "sheet_name"],
            "First row must be column headers",
        )
        self.assertEqual(
            table[1],
            ["create_flow", "flow1"],
            "Subsequent rows must contain values",
        )

    def test_values_must_be_strings(self):
        data = [
            {
                "boolean": True,
                "float": 1.23,
                "integer": "123",
                "string": "hello",
            },
        ]

        table = tabulate(data)

        self.assertEqual(table[1], ["true", "1.23", "123", "hello"])

    def test_columns_can_be_ordered_by_metadata(self):
        meta = {"headers": ["integer", "float", "string", "boolean"]}
        data = [
            {
                "boolean": True,
                "float": 1.23,
                "integer": "123",
                "string": "hello",
            },
        ]

        table = tabulate(data, meta)

        self.assertEqual(
            table[1],
            ["123", "1.23", "hello", "true"],
            "Columns should be in the same order as the headers metadata",
        )

    def test_arrays_use_single_cell_layout_by_default(self):
        data = [
            {"h1": ["yes", "no", 1, False]},
            {"h1": ("yes", "no", 1, False)},
        ]

        table = tabulate(data)

        self.assertEqual(table[1], ["yes | no | 1 | false"])
        self.assertEqual(table[2], ["yes | no | 1 | false"])

    def test_array_delimiters_are_escaped(self):
        data = [
            {"h1": [[1, 2], "3 | 4", "5 ; 6"]},
        ]

        table = tabulate(data)

        self.assertEqual(table[1], [r"1 ; 2 | 3 \| 4 | 5 \; 6"])

    def test_delimiters_in_templates_are_not_escaped(self):
        data = [
            {"h1": '{@ values | map(attribute="ID") @}'},
        ]

        table = tabulate(data)

        self.assertEqual(table[1], ['{@ values | map(attribute="ID") @}'])

    def test_single_item_array(self):
        data = [{"k1": ["seq1v1"]}]

        table = tabulate(data)

        self.assertEqual(table[1][0], "seq1v1 |")

    def test_arrays_with_empty_single_item(self):
        data = [{"k1": [""]}]

        table = tabulate(data)

        self.assertEqual(table[1][0], " |")

    def test_arrays_with_empty_last_item(self):
        data = [{"k1": ["v1", ""]}]

        table = tabulate(data)

        self.assertEqual(table[1][0], "v1 | |")

    def test_nested_arrays_within_a_single_cell(self):
        data = [
            {"k1": ["seq1v1", ["seq2v1", "seq2v2"]]},
        ]

        table = tabulate(data)

        self.assertEqual(table[1][0], "seq1v1 | seq2v1 ; seq2v2")

    def test_raise_exception_if_too_much_nesting_for_a_single_cell(self):
        data = [
            {"k1": ["seq1v1", ["seq2v1", ["seq3v1"]]]},
        ]

        self.assertRaises(Exception, tabulate, data)

    def test_arrays_use_wide_layout_if_indicated_by_metadata(self):
        meta = {
            "headers": [
                "choices",
                "choices",
                "choices",
                "choices",
            ]
        }
        data = [
            {
                "choices": ["yes", "no", 1, False],
            },
        ]

        table = tabulate(data, meta)

        self.assertEqual(table[0], ["choices", "choices", "choices", "choices"])
        self.assertEqual(table[1], ["yes", "no", "1", "false"])

    def test_objects_use_single_cell_layout_by_default(self):
        data = [
            {
                "obj": {
                    "prop1": "val1",
                    "prop2": "val2",
                },
            },
        ]

        table = tabulate(data)

        self.assertEqual(table[1], ["prop1; val1 | prop2; val2"])

    def test_object_with_single_property_within_cell_has_trailing_delimiter(self):
        data = [{"obj": {"k": "v"}}]

        table = tabulate(data)

        self.assertEqual(table[1], ["k; v |"])

    def test_objects_use_wide_layout_if_indicated_by_metadata(self):
        meta = {"headers": ["obj1.k1", "obj1.k2", "seq1.1.k1", "seq1.2.k2"]}
        data = [
            {
                "obj1": {
                    "k1": "obj1_k1_v",
                    "k2": "obj1_k2_v",
                },
                "seq1": [
                    {"k1": "seq1_k1_v"},
                    {"k2": "seq1_k2_v"},
                ],
            },
        ]

        table = tabulate(data, meta)

        self.assertEqual(
            table[0],
            ["obj1.k1", "obj1.k2", "seq1.1.k1", "seq1.2.k2"],
        )
        self.assertEqual(
            table[1],
            ["obj1_k1_v", "obj1_k2_v", "seq1_k1_v", "seq1_k2_v"],
        )


class TestUniversalToWorkbook(TestCase):
    def test_assembly(self):
        data = {
            "group1": [{"a": "a1", "b": "b1"}],
            "group2": [{"A": "A1", "B": "B1"}],
            "_idems": {
                "tabulate": {
                    "group1": {"headers": ["a", "b"]},
                    "group2": {"headers": ["B", "A"]},
                },
            },
        }

        workbook = bookify(data)

        self.assertEqual(len(workbook), 2)
        self.assertEqual(workbook[0][0], "group1")
        self.assertEqual(workbook[0][1], [["a", "b"], ["a1", "b1"]])
        self.assertEqual(workbook[1][0], "group2")
        self.assertEqual(
            workbook[1][1],
            [["B", "A"], ["B1", "A1"]],
            "Columns should be ordered according to metadata",
        )
        self.assertEqual(
            data,
            {
                "group1": [{"a": "a1", "b": "b1"}],
                "group2": [{"A": "A1", "B": "B1"}],
                "_idems": {
                    "tabulate": {
                        "group1": {"headers": ["a", "b"]},
                        "group2": {"headers": ["B", "A"]},
                    },
                },
            },
            "Input data should not be mutated",
        )


class TestConvertWorkbookToUniversal(TestCase):

    def test_workbook_converts_to_object(self):
        workbook = DatasetSheetReader(
            [
                Dataset(("t1a1", "t1b1"), headers=("T1A", "T1B"), title="table1"),
                Dataset(("t2a1", "t2b1"), headers=("T2A", "T2B"), title="table2"),
            ],
            "test"
        )

        nested = parse_tables(workbook)

        self.assertIsInstance(nested, dict)
        self.assertEqual(list(nested.keys()), ["_idems", "table1", "table2"])
        self.assertEqual(
            list(nested["_idems"]["tabulate"].keys()),
            ["table1", "table2"],
        )


class TestConvertTableToNested(TestCase):

    def test_default_type_is_string(self):
        self.assertEqual(
            parse_table(
                title="title",
                headers=["a"],
                rows=[["a1"]],
            ),
            {
                "_idems": {"tabulate": {"title": {"headers": ["a"]}}},
                "title": [{"a": "a1"}],
            },
        )

    def test_table_must_have_title(self):
        self.assertEqual(parse_table(), {"table": []})

    def test_integer_as_string_is_int(self):
        parsed = parse_table(headers=["a"], rows=[["123"]])

        self.assertEqual(parsed["table"][0]["a"], 123)

    def test_boolean_as_string_is_bool(self):
        parsed = parse_table(headers=("a", "b"), rows=[("true", "false")])

        self.assertEqual(parsed["table"][0]["a"], True)
        self.assertEqual(parsed["table"][0]["b"], False)

    def test_delimited_string_is_array(self):
        parsed = parse_table(headers=["a"], rows=[["one | 2 | true | 3.4"]])

        self.assertEqual(parsed["table"][0]["a"], ["one", 2, True, 3.4])

    def test_columns_with_same_name_are_grouped_into_list(self):
        parsed = parse_table(headers=["a"] * 4, rows=[("one", "2", "true", "3.4")])

        self.assertEqual(parsed["table"][0]["a"], ["one", 2, True, 3.4])

    def test_columns_with_same_name_and_delimited_strings_is_2d_array(self):
        parsed = parse_table(headers=["a"] * 2, rows=[("one | 2", "true | 3.4")])

        self.assertEqual(parsed["table"][0]["a"], [["one", 2], [True, 3.4]])

    def test_column_using_dot_notation_is_nested_object_property(self):
        parsed = parse_table(
            headers=("obj.prop1", "obj.prop2"),
            rows=[("one", "2")],
        )

        self.assertEqual(parsed["table"][0]["obj"], {"prop1": "one", "prop2": 2})
        self.assertEqual(
            parsed["_idems"]["tabulate"]["table"]["headers"],
            ("obj.prop1", "obj.prop2"),
        )

    def test_nested_object_with_2d_array_property_value(self):
        parsed = parse_table(headers=["obj.k1"] * 2, rows=[["1 | 2", "3 | 4"]])

        self.assertEqual(parsed["table"][0]["obj"], {"k1": [[1, 2], [3, 4]]})

    def test_nested_object_with_nested_object(self):
        parsed = parse_table(
            headers=["obj.k1"] * 2,
            rows=[["k2; 2 | k3; false", "k4; v4 | k5; true"]],
        )

        self.assertEqual(
            parsed["table"][0]["obj"],
            {"k1": [[["k2", 2], ["k3", False]], [["k4", "v4"], ["k5", True]]]},
        )


class TestCellConversion(TestCase):

    def setUp(self):
        self.func = parse_cell

    def test_convert_cell_string_to_number(self):
        self.assertEqual(self.func("123"), 123)
        self.assertEqual(self.func("1.23"), 1.23)

    def test_output_clean_string_if_no_conversion_possible(self):
        self.assertEqual(self.func("one"), "one")
        self.assertEqual(self.func(" one "), "one")
        self.assertEqual(self.func(""), "")
        self.assertEqual(self.func("http://example.com/"), "http://example.com/")
        self.assertEqual(self.func("k1: v1"), "k1: v1")

    def test_raises_error_if_not_string_input(self):
        self.assertRaises(TypeError, self.func, None)
        self.assertRaises(TypeError, self.func, 123)

    def test_convert_cell_string_to_bool(self):
        self.assertEqual(self.func("true"), True)
        self.assertEqual(self.func(" true "), True)
        self.assertEqual(self.func("false"), False)

    def test_convert_cell_string_to_list(self):
        self.assertEqual(self.func("one | 2 | false"), ["one", 2, False])
        self.assertEqual(self.func("one ; 2 ; false"), ["one", 2, False])
        self.assertEqual(self.func("one |"), ["one"])
        self.assertEqual(self.func("|"), [""])
        self.assertEqual(self.func("| 2 |"), ["", 2])
        self.assertEqual(self.func("a||"), ["a", ""])
        self.assertEqual(self.func("k1 | v1 : k2 | v2"), ["k1", "v1 : k2", "v2"])

    def test_convert_cell_string_to_list_of_lists(self):
        self.assertEqual(self.func("k1; v1 |"), [["k1", "v1"]])
        self.assertEqual(self.func("k1; k2; v2 |"), [["k1", "k2", "v2"]])
        self.assertEqual(self.func("k1; 1 | k2; true"), [["k1", 1], ["k2", True]])

    def test_delimiters_can_be_configured(self):
        self.assertEqual(
            self.func(
                "click | auth: sign_in_google; click | emit: force_reprocess",
                delimiters=";|:",
            ),
            [
                ["click", ["auth", "sign_in_google"]],
                ["click", ["emit", "force_reprocess"]],
            ],
        )

    def test_inline_templates_are_preserved(self):
        self.assertEqual(self.func("{{ template }}"), "{{ template }}")
        self.assertEqual(self.func("{@ template @}"), "{@ template @}")
        self.assertEqual(
            self.func("{% if other_option!=" "%}1wc;1wt;1wb{%endif-%}"),
            "{% if other_option!=" "%}1wc;1wt;1wb{%endif-%}",
        )
        self.assertEqual(self.func("{{ template }} |"), "{{ template }} |")
        self.assertEqual(
            self.func("{{ template }} | something | {{ blah }}"),
            "{{ template }} | something | {{ blah }}",
        )
        self.assertEqual(
            self.func(
                "{{3*(steps.values()|length -1)}}|{{3*(steps.values()|length -1)+2}}"
            ),
            "{{3*(steps.values()|length -1)}}|{{3*(steps.values()|length -1)+2}}",
        )
        self.assertEqual(
            self.func("6;0{%if skip_option != " " -%};skip{% endif %}"),
            "6;0{%if skip_option != " " -%};skip{% endif %}",
        )
        self.assertEqual(
            self.func('@(fields.survey_behave & "no|")'),
            '@(fields.survey_behave & "no|")',
        )

    def test_delimiters_can_be_escaped(self):
        self.assertEqual(
            self.func(r"1 ; 2 | 3 \| 4 | 5 \; 6 \|"),
            [[1, 2], "3 | 4", "5 ; 6 |"],
        )



import copy

from rpft.rapidpro.models.common import (
    FlowReference,
    Group,
    mangle_string,
    UserContactField,
)
from rpft.rapidpro.models.exceptions import RapidProActionError
from rpft.rapidpro.utils import generate_new_uuid
from rpft.parsers.creation.flowrowmodel import dict_to_list_of_pairs

# TODO: Check enter flow
# Node classification:
# - Action-only node (for various actions)
# - No action, split by variable [this includes wait_for_response]
# - Action + split by variable:
# - Enter flow (Router with Completed/Expired)
# - Call webhook (Router with Success/Failure)
# - No action, split by random


class Action:
    def from_dict(data):
        # Create a generic Action, and cast it to the specific Action subclass
        # in order to bypass the constructor of the subclass
        if "type" not in data:
            raise RapidProActionError("RapidProAction must have a type.")
        action_type = data["type"]
        # TODO: Can we make this more smooth by invoking subclass constructors?
        # And make the Action class abstract?
        action = Action(action_type)
        cls = action_map[action_type]
        action.__class__ = cls
        # Fill in the fields of the object
        action._assign_fields_from_dict(data)
        return action

    def _assign_fields_from_dict(self, data):
        self.__dict__ = copy.deepcopy(data)

    def __init__(self, type, **kwargs):
        self.uuid = generate_new_uuid()
        self.type = type
        for k, v in kwargs.items():
            setattr(self, k, v)

    def record_global_uuids(self, uuid_dict):
        pass

    def assign_global_uuids(self, uuid_dict):
        pass

    def render(self):
        return {
            "uuid": self.uuid,
            "type": self.type,
        }

    def short_name(self):
        short_type = short_types.get(self.type, self.type)
        short_value = mangle_string(self.main_value())
        return f"{short_type}.{short_value}"

    def main_value(self):
        raise NotImplementedError

    def get_row_model_fields(self):
        # This should probably be an abstract method returning a partially
        # instantiated row model.
        raise NotImplementedError


class DefaultRenderedAction(Action):
    def render(self):
        return self.__dict__

    def get_row_model_fields(self):
        return NotImplementedError


class AddContactURNAction(DefaultRenderedAction):
    def __init__(self, **kwargs):
        super().__init__("add_contact_urn", **kwargs)

    def main_value(self):
        return self.path

    def get_row_model_fields(self):
        return {
            "type": self.type,
            "mainarg_value": self.path,
            "urn_scheme": self.scheme if self.scheme != "tel" else "",
        }


class CallWebhookAction(DefaultRenderedAction):
    def __init__(self, **kwargs):
        super().__init__("call_webhook", **kwargs)

    def main_value(self):
        return self.body

    def get_row_model_fields(self):
        headers = dict_to_list_of_pairs(self.headers)
        return {
            "type": self.type,
            "webhook": {
                "body": self.body,
                "url": self.url,
                "headers": headers,
                "method": self.method,
            },
            "save_name": self.result_name,
        }


class TransferAirtimeAction(DefaultRenderedAction):
    def __init__(self, **kwargs):
        assert "amounts" in kwargs
        super().__init__("transfer_airtime", **kwargs)

    def main_value(self):
        return self.amounts

    def get_row_model_fields(self):
        amounts = {k: str(v) for k, v in self.amounts.items()}
        amounts = dict_to_list_of_pairs(amounts)
        return {
            "type": self.type,
            "mainarg_dict": amounts,
            "save_name": self.result_name,
        }


class WhatsAppMessageTemplating:
    def __init__(self, name, template_uuid, variables, uuid=None):
        self.name = name
        self.uuid = uuid or generate_new_uuid()
        self.template_uuid = template_uuid
        self.variables = variables

    def from_whats_app_templating_model(model):
        return WhatsAppMessageTemplating(
            name=model.name,
            template_uuid=model.uuid,
            variables=model.variables,
        )

    def from_rapid_pro_templating(templating):
        return WhatsAppMessageTemplating(
            templating["template"]["name"],
            templating["template"]["uuid"],
            templating["variables"],
            templating["uuid"],
        )

    def to_whats_app_templating_dict(self):
        return {
            "name": self.name,
            "uuid": self.template_uuid,
            "variables": self.variables,
        }

    def render(self):
        return {
            "template": {
                "name": self.name,
                "uuid": self.template_uuid,
            },
            "uuid": self.uuid,
            "variables": self.variables,
        }


class SendMessageAction(Action):
    def __init__(
        self, text, attachments=None, quick_replies=None, all_urns=None, templating=None
    ):
        super().__init__("send_msg")
        if not text:
            raise RapidProActionError("send_msg action requires non-empty text.")
        self.text = text
        self.attachments = attachments or list()
        self.quick_replies = quick_replies or list()
        self.all_urns = all_urns
        self.templating = templating

    def _assign_fields_from_dict(self, data):
        data_copy = copy.deepcopy(data)
        if "templating" in data:
            templating = data_copy.pop("templating")
        super()._assign_fields_from_dict(data_copy)
        if "templating" in data:
            self.templating = WhatsAppMessageTemplating.from_rapid_pro_templating(
                templating
            )

    def add_attachment(self, attachment):
        self.attachments.append(attachment)

    def add_quick_reply(self, quick_reply):
        self.quick_replies.append(quick_reply)

    def main_value(self):
        return self.text

    def _get_attachments(self):
        return [attachment for attachment in self.attachments if attachment]

    def render(self):
        # Can we find a more compact way of invoking the superclass
        # to render the common fields?
        render_dict = super().render()
        render_dict.update(
            {
                "text": self.text,
                "attachments": self._get_attachments(),
                "quick_replies": self.quick_replies,
            }
        )

        # Refactor this into a method to avoid code replication
        if hasattr(self, "all_urns") and self.all_urns:
            render_dict.update({"all_urns": self.all_urns})
        if hasattr(self, "topic") and self.topic:
            render_dict.update({"topic": self.topic})
        if hasattr(self, "templating") and self.templating:
            render_dict.update({"templating": self.templating.render()})

        return render_dict

    def get_row_model_fields(self):
        # TODO: templating
        attachment_by_type = {}
        attachments = self._get_attachments()
        # If there are more than 1 attachment, we cannot encode their
        # order if we use the image/audio/video column.
        # Thus we only use these if there is exactly one attachment,
        # and otherwise we use the general attachment list.
        if len(attachments) == 1:
            attachment = attachments[0]
            for attachment_type in ["image", "audio", "video"]:
                if attachment.startswith(f"{attachment_type}:"):
                    attachment_by_type[attachment_type] = attachment[6:]
                    attachments = []
                    break

        data = {
            "type": "send_message",
            "mainarg_message_text": self.text,
            "choices": self.quick_replies,
            "image": attachment_by_type.get("image", ""),
            "audio": attachment_by_type.get("audio", ""),
            "video": attachment_by_type.get("video", ""),
            "attachments": attachments,
        }

        if hasattr(self, "templating") and self.templating:
            data["wa_template"] = (
                WhatsAppMessageTemplating.to_whats_app_templating_dict(self.templating)
            )

        return data


class SetContactFieldAction(Action):

    VALUE_LENGTH_LIMIT = 640

    def __init__(self, field_name, value):
        super().__init__("set_contact_field")
        self.field = UserContactField(field_name)
        self.value = value

        if len(value) > SetContactFieldAction.VALUE_LENGTH_LIMIT:
            raise RapidProActionError(
                "Contact field value length limit exceeded",
                {
                    "length": len(value),
                    "limit": SetContactFieldAction.VALUE_LENGTH_LIMIT,
                    "name": field_name,
                    "value": value,
                },
            )

    def _assign_fields_from_dict(self, data):
        assert "field" in data
        data_copy = copy.deepcopy(data)
        data_copy["field"] = UserContactField(**data_copy["field"])
        super()._assign_fields_from_dict(data_copy)

    def main_value(self):
        return self.field.key

    def render(self):
        return {
            "uuid": self.uuid,
            "type": self.type,
            "field": self.field.render(),
            "value": self.value,
        }

    def get_row_model_fields(self):
        return {
            "type": "save_value",
            "mainarg_value": self.value,
            "save_name": self.field.key,
        }


# This action captures the following action types:
# set_contact_channel
# set_contact_language
# set_contact_name
# set_contact_status
# set_contact_timezone
class SetContactPropertyAction(Action):
    def __init__(self, property, value):
        super().__init__(f"set_contact_{property}")
        self.property = property
        if not value:
            raise RapidProActionError(
                f"{property} must be non-empty for set_contact_{property}."
            )
        self.value = value

    def _assign_fields_from_dict(self, data):
        assert "type" in data
        action_type = data["type"]
        assert action_type.find("set_contact_") != -1
        property = action_type.replace("set_contact_", "")
        assert property in data
        assert property in ["channel", "language", "name", "status", "timezone"]
        data_copy = copy.deepcopy(data)
        super()._assign_fields_from_dict(data_copy)
        self.property = property
        self.value = data_copy.pop(property)

    def main_value(self):
        return self.property

    def render(self):
        return {"uuid": self.uuid, "type": self.type, self.property: self.value}

    def get_row_model_fields(self):
        return {
            "type": self.type,
            "mainarg_value": self.value,
        }


class GenericGroupAction(Action):
    def __init__(self, type, groups):
        super().__init__(type)
        self.groups = groups

    def _assign_fields_from_dict(self, data):
        assert "groups" in data
        groups = []
        for group in data["groups"]:
            groups.append(Group.from_dict(group))
        data = copy.deepcopy(data)  # don't mutate the input
        data["groups"] = groups
        super()._assign_fields_from_dict(data)

    def record_global_uuids(self, uuid_dict):
        for group in self.groups:
            group.record_uuid(uuid_dict)

    def assign_global_uuids(self, uuid_dict):
        for group in self.groups:
            group.assign_uuid(uuid_dict)

    def main_value(self):
        return self.groups[0].name if self.groups else ""

    def render(self):
        return NotImplementedError

    def get_row_model_fields(self):
        # abstract method
        obj_id = [group.uuid for group in self.groups][0] if self.groups else ""
        return {
            "mainarg_groups": [group.name for group in self.groups],
            "obj_id": obj_id or "",  # 0th element as obj_id is not yet a list.
        }


class AddContactGroupAction(GenericGroupAction):
    def __init__(self, groups):
        super().__init__("add_contact_groups", groups)

    def add_group(self, group):
        self.groups.append(group)

    def render(self):
        return {
            "type": self.type,
            "uuid": self.uuid,
            "groups": [group.render() for group in self.groups],
        }

    def get_row_model_fields(self):
        fields = super().get_row_model_fields()
        fields["type"] = "add_to_group"
        return fields


class RemoveContactGroupAction(GenericGroupAction):
    def __init__(self, groups, all_groups=None):
        super().__init__("remove_contact_groups", groups)
        self.all_groups = all_groups

    def render(self):
        render_dict = {
            "type": self.type,
            "uuid": self.uuid,
            "groups": [group.render() for group in self.groups],
        }
        if hasattr(self, "all_groups") and self.all_groups:
            render_dict.update({"all_groups": self.all_groups})
        return render_dict

    def get_row_model_fields(self):
        fields = super().get_row_model_fields()
        fields["type"] = "remove_from_group"
        return fields


class SetRunResultAction(Action):
    def __init__(self, name, value, category=""):
        super().__init__("set_run_result")
        self.name = name
        self.value = value
        self.category = category
        if len(value) > 640:
            raise RapidProActionError(
                "Flow results are limited to 640 characters, but value has length"
                f" {len(value)}"
            )

    def main_value(self):
        return self.name

    def _assign_fields_from_dict(self, data):
        assert "name" in data
        assert "value" in data
        super()._assign_fields_from_dict(data)
        if "category" not in data:
            self.category = ""

    def render(self):
        render_dict = {
            "type": self.type,
            "name": self.name,
            "value": self.value,
            "uuid": self.uuid,
        }
        if self.category:
            render_dict.update(
                {
                    "category": self.category,
                }
            )
        return render_dict

    def get_row_model_fields(self):
        output_dict = {
            "type": "save_flow_result",
            "mainarg_value": self.value,
            "save_name": self.name,
        }
        if self.category:
            output_dict.update(
                {
                    "result_category": self.category,
                }
            )
        return output_dict


class EnterFlowAction(Action):
    def __init__(self, flow_name, flow_uuid=None):
        super().__init__("enter_flow")
        self.flow = FlowReference(flow_name, flow_uuid)

    def _assign_fields_from_dict(self, data):
        assert "flow" in data
        data = copy.deepcopy(data)  # don't mutate the input
        data["flow"] = FlowReference.from_dict(data["flow"])
        super()._assign_fields_from_dict(data)

    def record_global_uuids(self, uuid_dict):
        self.flow.record_uuid(uuid_dict)

    def assign_global_uuids(self, uuid_dict):
        self.flow.assign_uuid(uuid_dict)

    def main_value(self):
        return self.flow.name

    def render(self):
        return {"type": self.type, "uuid": self.uuid, "flow": self.flow.render()}

    def get_row_model_fields(self):
        return {
            "type": "start_new_flow",
            "mainarg_flow_name": self.flow.name,
            "obj_id": self.flow.uuid or "",
        }


action_map = {
    "add_contact_groups": AddContactGroupAction,
    "add_contact_urn": AddContactURNAction,
    "add_input_labels": DefaultRenderedAction,
    "call_classifier": DefaultRenderedAction,
    "call_resthook": DefaultRenderedAction,
    "call_webhook": CallWebhookAction,
    "enter_flow": EnterFlowAction,
    "open_ticket": DefaultRenderedAction,
    "play_audio": DefaultRenderedAction,
    "remove_contact_groups": RemoveContactGroupAction,
    "say_msg": DefaultRenderedAction,
    "send_broadcast": DefaultRenderedAction,
    "send_email": DefaultRenderedAction,
    "send_msg": SendMessageAction,
    "set_contact_channel": SetContactPropertyAction,
    "set_contact_field": SetContactFieldAction,
    "set_contact_language": SetContactPropertyAction,
    "set_contact_name": SetContactPropertyAction,
    "set_contact_status": SetContactPropertyAction,
    "set_contact_timezone": SetContactPropertyAction,
    "set_run_result": SetRunResultAction,
    "start_session": DefaultRenderedAction,
    "transfer_airtime": TransferAirtimeAction,
}

short_types = {
    "call_webhook": "webhook",
    "enter_flow": "flow",
    "send_msg": "msg",
}



from unittest import TestCase

from rpft.parsers.creation.contentindexparser import (
    ContentIndexParser,
    DataSheet,
)
from rpft.parsers.creation.models import (
    Assignment,
    Condition,
    ConditionWithMessage,
    Message,
    PostProcessing,
    SurveyConfig,
)
from rpft.parsers.creation.globalrowmodels import SurveyQuestionRowModel
from rpft.parsers.creation.surveyparser import (
    apply_to_all_str,
    Survey,
    SurveyQuestion,
)
from rpft.parsers.sheets import CSVSheetReader
from rpft.rapidpro.simulation import Context, traverse_flow
from rpft.sources import SheetDataSource

from tests import TESTS_ROOT
from tests.mocks import MockSheetReader
from tests.utils import csv_join


class TestTemplate(TestCase):
    def assertFlowMessages(self, render_output, flow_name, actions_exp, context=None):
        flows = [flow for flow in render_output["flows"] if flow["name"] == flow_name]

        self.assertTrue(
            len(flows) > 0,
            msg=f'Flow with name "{flow_name}" does not exist in output.',
        )

        actions = traverse_flow(flows[0], context or Context())

        self.assertEqual(actions, actions_exp)


class TestSurveyParser(TestTemplate):
    def test_basic_global_model(self):
        ci_sheet = csv_join(
            "type,sheet_name,data_model",
            "data_sheet,survey_data,SurveyQuestionRowModel",
        )
        survey_data = csv_join(
            "ID,type,question",
            "name,text,Enter your name",
            "else,text,Enter something else",
        )
        definition = ContentIndexParser(
            SheetDataSource([MockSheetReader(ci_sheet, {"survey_data": survey_data})])
        ).definition
        datamodelA = definition.get_data_sheet_row("survey_data", "name")
        datamodelB = definition.get_data_sheet_row("survey_data", "else")

        self.assertEqual(datamodelA.type, "text")
        self.assertEqual(datamodelA.messages[0].text, "Enter your name")
        self.assertEqual(datamodelB.type, "text")
        self.assertEqual(datamodelB.messages[0].text, "Enter something else")

    def test_basic_survey(self):
        ci_sheet = csv_join(
            "type,sheet_name,data_sheet,new_name,data_model",
            "data_sheet,survey_data,,,SurveyQuestionRowModel",
            "survey,,survey_data,Survey Name,",
        )
        survey_data = csv_join(
            "ID,type,question,variable,completion_variable,expiration.message",
            "first,text,First question?,first,first_complete,",
            "second,text,Second question?,second,second_complete,You waited too long",
            "third,text,Third question?,,,",
        )

        output = (
            ContentIndexParser(
                SheetDataSource(
                    [
                        CSVSheetReader(TESTS_ROOT / "input/survey_templates"),
                        MockSheetReader(ci_sheet, {"survey_data": survey_data}),
                    ],
                )
            )
            .parse_all()
            .render()
        )

        self.assertFlowMessages(
            output,
            "survey - Survey Name - question - first",
            [
                ("set_run_result", "dummy"),
                ("send_msg", "First question?"),
                ("set_contact_field", "first"),
                ("set_contact_field", "first_complete"),
            ],
            Context(inputs=["First answer"]),
        )

        self.assertFlowMessages(
            output,
            "survey - Survey Name - question - third",
            [
                ("set_run_result", "dummy"),
                ("send_msg", "Third question?"),
                ("set_contact_field", "sq_surveyname_third"),
                ("set_contact_field", "sq_surveyname_third_complete"),
            ],
            Context(inputs=["Third answer"]),
        )

        self.assertFlowMessages(
            output,
            "survey - Survey Name",
            [
                ("enter_flow", "survey - Survey Name - question - first"),
                ("enter_flow", "survey - Survey Name - question - second"),
                ("send_msg", "You waited too long"),
                ("set_run_result", "expired"),
            ],
            Context(inputs=["completed", "expired"]),
        )

        self.assertFlowMessages(
            output,
            "survey - Survey Name",
            [
                ("enter_flow", "survey - Survey Name - question - first"),
                ("enter_flow", "survey - Survey Name - question - second"),
                ("enter_flow", "survey - Survey Name - question - third"),
                ("set_run_result", "proceed"),
            ],
            Context(inputs=["completed", "completed", "completed"]),
        )

        self.assertFlowMessages(
            output,
            "survey - Survey Name",
            [
                ("enter_flow", "survey - Survey Name - question - first"),
            ],
            Context(inputs=["completed"], variables={"@child.results.stop": "yes"}),
        )

    def test_stop_condition(self):
        ci_sheet = csv_join(
            "type,sheet_name,data_sheet,data_row_id,new_name,data_model",
            "data_sheet,survey_data,,,,SurveyQuestionRowModel",
            "survey,,survey_data,,Survey Name,",
        )
        survey_data = csv_join(
            "ID,type,question,variable,completion_variable,stop.conditions.1.condition,stop.conditions.1.message,stop.conditions.2.condition,stop.conditions.2.message",  # noqa: E501
            "age,text,Enter your age,,,18|@answer|has_number_lt|,You are too young,25|@answer|has_number_gt|,You are too old",  # noqa: E501
            "name,text,Enter your name,,,,",
        )

        output = (
            ContentIndexParser(
                SheetDataSource(
                    [
                        CSVSheetReader(TESTS_ROOT / "input/survey_templates"),
                        MockSheetReader(ci_sheet, {"survey_data": survey_data}),
                    ]
                )
            )
            .parse_all()
            .render()
        )

        self.assertFlowMessages(
            output,
            "survey - Survey Name - question - age",
            [
                ("set_run_result", "dummy"),
                ("send_msg", "Enter your age"),
                ("set_contact_field", "sq_surveyname_age"),
                ("set_contact_field", "sq_surveyname_age_complete"),
                ("set_run_result", "stop"),
                ("send_msg", "You are too young"),
            ],
            Context(inputs=["15"], variables={"@fields.sq_surveyname_age": "15"}),
        )

        self.assertFlowMessages(
            output,
            "survey - Survey Name - question - name",
            [
                ("set_run_result", "dummy"),
                ("send_msg", "Enter your name"),
                ("set_contact_field", "sq_surveyname_name"),
                ("set_contact_field", "sq_surveyname_name_complete"),
            ],
            Context(inputs=["21"]),
        )

        self.assertFlowMessages(
            output,
            "survey - Survey Name - question - age",
            [
                ("set_run_result", "dummy"),
                ("send_msg", "Enter your age"),
                ("set_contact_field", "sq_surveyname_age"),
                ("set_contact_field", "sq_surveyname_age_complete"),
                ("set_run_result", "stop"),
                ("send_msg", "You are too old"),
            ],
            Context(inputs=["30"], variables={"@fields.sq_surveyname_age": "30"}),
        )

        self.assertFlowMessages(
            output,
            "survey - Survey Name - question - name",
            [
                ("set_run_result", "dummy"),
                ("send_msg", "Enter your name"),
                ("set_contact_field", "sq_surveyname_name"),
                ("set_contact_field", "sq_surveyname_name_complete"),
            ],
            Context(inputs=["My name"]),
        )

    def test_template_overwrite(self):
        ci_sheet = csv_join(
            "type,sheet_name,data_sheet,new_name,data_model,template_arguments",
            "data_sheet,survey_data,,,SurveyQuestionRowModel,",
            "template_definition,template_survey_wrapper,,,,end_flow",
            "template_definition,template_survey_question_type_text,,,,",
            "survey,,survey_data,Survey Name,,my_end_flow",
        )
        survey_data = csv_join(
            "ID,type,question",
            "name,text,Enter your name",
        )
        template_survey_wrapper = csv_join(
            "type,from,message_text,condition",
            "send_message,start,Here's the survey,",
            "start_new_flow,,survey - {{survey_name}} - question - {{questions.0.ID}},",
            "start_new_flow,,{{end_flow}},Completed",
        )
        template_survey_question_type_text = csv_join(
            "row_id,type,from,message_text,save_name\n",
            ",send_message,start,Here's the question,\n",
            ",send_message,,{{messages.0.text}},\n",
            ",wait_for_response,,,input\n",
        )
        sheet_dict = {
            "survey_data": survey_data,
            "template_survey_wrapper": template_survey_wrapper,
            "template_survey_question_type_text": template_survey_question_type_text,
        }
        output = (
            ContentIndexParser(
                SheetDataSource(
                    [
                        CSVSheetReader(TESTS_ROOT / "input/survey_templates"),
                        MockSheetReader(ci_sheet, sheet_dict),
                    ]
                )
            )
            .parse_all()
            .render()
        )

        self.assertFlowMessages(
            output,
            "survey - Survey Name - question - name",
            [
                ("send_msg", "Here's the question"),
                ("send_msg", "Enter your name"),
                ("set_contact_field", "sq_surveyname_name"),
                ("set_contact_field", "sq_surveyname_name_complete"),
            ],
            Context(inputs=["My name"]),
        )

        self.assertFlowMessages(
            output,
            "survey - Survey Name",
            [
                ("send_msg", "Here's the survey"),
                ("enter_flow", "survey - Survey Name - question - name"),
                ("enter_flow", "my_end_flow"),
            ],
            Context(inputs=["completed", "completed"]),
        )

    def test_template_arguments(self):
        ci_sheet = csv_join(
            "type,sheet_name,data_sheet,new_name,data_model,template_arguments",
            "data_sheet,survey_data,,,SurveyQuestionRowModel,",
            "survey,,survey_data,Survey Name,,survey_defaults",
        )
        survey_data = csv_join(
            "ID,type,question,variable,completion_variable",
            "first,text,First question?,first,",
            "second,text,Second question?,second,second_complete",
        )

        output = (
            ContentIndexParser(
                SheetDataSource(
                    [
                        CSVSheetReader(
                            TESTS_ROOT / "input/survey_templates_using_defaults"
                        ),
                        MockSheetReader(ci_sheet, {"survey_data": survey_data}),
                    ],
                )
            )
            .parse_all()
            .render()
        )

        self.assertFlowMessages(
            output,
            "survey - Survey Name - question - first",
            [
                ("set_run_result", "dummy"),
                ("send_msg", "First question?"),
                ("send_msg", "Question level message"),
                ("set_contact_field", "first"),
                ("set_contact_field", "first_complete"),
            ],
            Context(inputs=["First answer"]),
        )

        self.assertFlowMessages(
            output,
            "survey - Survey Name",
            [
                ("enter_flow", "survey - Survey Name - question - first"),
                ("enter_flow", "survey - Survey Name - question - second"),
                ("send_msg", "Survey level message"),
            ],
            Context(inputs=["completed", "completed"]),
        )

    def test_create_question(self):
        ci_sheet = csv_join(
            "type,sheet_name,data_sheet,new_name,data_model,template_arguments,data_row_id",  # noqa: E501
            "data_sheet,survey_data,,,SurveyQuestionRowModel,,",
            "survey_question,,survey_data,Survey Name,,survey_defaults,first1",
            "survey_question,,survey_data,Survey Name,,survey_defaults,second2",
            "survey_question,,survey_data,Survey Name,,survey_defaults,third3",
        )
        survey_data = csv_join(
            "ID,type,question,variable,completion_variable",
            "first1,text,First question?,,",
            "second2,text,Second question?,second,",
            "third3,text,@answer question?,third,third_complete",
        )

        output = (
            ContentIndexParser(
                SheetDataSource(
                    [
                        CSVSheetReader(
                            TESTS_ROOT / "input/survey_templates_using_defaults"
                        ),
                        MockSheetReader(ci_sheet, {"survey_data": survey_data}),
                    ],
                )
            )
            .parse_all()
            .render()
        )

        self.assertFlowMessages(
            output,
            "survey - Survey Name - question - first1",
            [
                ("set_run_result", "dummy"),
                ("send_msg", "First question?"),
                ("send_msg", "Question level message"),
                ("set_contact_field", "sq_surveyname_first1"),
                ("set_contact_field", "sq_surveyname_first1_complete"),
            ],
            Context(inputs=["First answer"]),
        )

        self.assertFlowMessages(
            output,
            "survey - Survey Name - question - second2",
            [
                ("set_run_result", "dummy"),
                ("send_msg", "Second question?"),
                ("send_msg", "Question level message"),
                ("set_contact_field", "second"),
                ("set_contact_field", "second_complete"),
            ],
            Context(inputs=["Second answer"]),
        )

        self.assertFlowMessages(
            output,
            "survey - Survey Name - question - third3",
            [
                ("set_run_result", "dummy"),
                ("send_msg", "@fields.third question?"),
                ("send_msg", "Question level message"),
                ("set_contact_field", "third"),
                ("set_contact_field", "third_complete"),
            ],
            Context(inputs=["Third answer"]),
        )


class TestSurveyPreprocessing(TestCase):
    def test_apply_to_all_str(self):
        def f(s):
            return s.replace("a", "b")

        self.assertEqual(apply_to_all_str("a", f), "b")
        self.assertEqual(apply_to_all_str(["a", "name"], f), ["b", "nbme"])
        self.assertEqual(
            apply_to_all_str({"a": "name", "b": "ant"}, f), {"a": "nbme", "b": "bnt"}
        )
        self.assertEqual(
            apply_to_all_str({"a": "name", "b": "ant"}, f), {"a": "nbme", "b": "bnt"}
        )

        cond = ConditionWithMessage(
            condition=Condition(
                value="a",
                variable="abc",
            ),
            message="auto",
        )
        cond2 = ConditionWithMessage(
            condition=Condition(
                value="b",
                variable="bbc",
            ),
            message="buto",
        )
        self.assertEqual(apply_to_all_str(cond, f), cond2)

    def test_replacements(self):
        question2 = SurveyQuestionRowModel(
            ID="question2",
            type="qtype",
            variable="sq_question2",
            completion_variable="sq_question2_completion",
            messages=[
                Message(
                    text="Previously you answered @fields.sq_s1_question1. "
                    "Now answer Question 2:",
                )
            ],
            relevant=[
                Condition(
                    value="4",
                    variable="@fields.sq_s1_question1",
                ),
                Condition(
                    value="5",
                    variable="@fields.sq_s1_question1b",
                ),
                Condition(
                    value="6",
                    variable="@prefix_question0",
                ),
                Condition(
                    value="7",
                    variable="custom_@surveyid_@questionid",
                ),
            ],
            postprocessing=PostProcessing(
                assignments=[
                    Assignment(
                        variable="@answerid_bucket",
                        value="@answer",
                    ),
                    Assignment(
                        variable="new_variable2",
                        value="@fields.sq_s1_question1",
                    ),
                ],
                flow="my_flow",
            ),
        )

        question2_replaced = SurveyQuestionRowModel(
            ID="question2",
            type="qtype",
            variable="pre_sq_question2",
            completion_variable="pre_sq_question2_completion",
            messages=[
                Message(
                    text="Previously you answered @fields.pre_sq_s1_question1. "
                    "Now answer Question 2:",
                )
            ],
            relevant=[
                Condition(
                    value="4",
                    variable="@fields.pre_sq_s1_question1",
                ),
                Condition(
                    value="5",
                    variable="@fields.sq_s1_question1b",
                ),
                Condition(
                    value="6",
                    variable="@fields.sq_s1_question0",
                ),
                Condition(
                    value="7",
                    variable="custom_s1_question2",
                ),
            ],
            postprocessing=PostProcessing(
                assignments=[
                    Assignment(
                        variable="pre_sq_question2_bucket",
                        value="@fields.pre_sq_question2",
                    ),
                    Assignment(
                        variable="pre_new_variable2",
                        value="@fields.pre_sq_s1_question1",
                    ),
                ],
                flow="my_flow",
            ),
        )

        question2_copy = SurveyQuestion("Survey Name (unused)", question2)
        prefix = "pre_"
        question2_copy.apply_shorthand_substitutions("s1")
        question2_copy.apply_prefix_renaming(prefix)
        question2_copy.apply_prefix_substitutions(
            [
                "sq_s1_question1",
                "sq_s1_question1_complete",
                "sq_question2",
                "new_variable",
                "new_variable2",
            ],
            prefix,
        )

        self.assertEqual(question2_copy.data_row, question2_replaced)

        # Use this data in a survey
        question1 = SurveyQuestionRowModel(
            ID="question1",
            type="qtype",
            messages=[Message(text="Tell us the answer to Question 1")],
        )
        question1_replaced = SurveyQuestionRowModel(
            ID="question1",
            variable="pre_sq_s1_question1",
            completion_variable="pre_sq_s1_question1_complete",
            type="qtype",
            messages=[Message(text="Tell us the answer to Question 1")],
        )

        survey = Survey(
            "S 1~~",
            DataSheet(
                {"question1": question1, "question2": question2}, SurveyQuestionRowModel
            ),
            SurveyConfig(variable_prefix="pre_"),
        )
        survey.preprocess_data_rows()
        self.assertEqual(
            survey.get_question_by_id("question1").data_row, question1_replaced
        )
        self.assertEqual(
            survey.get_question_by_id("question2").data_row, question2_replaced
        )



import unittest

from rpft.rapidpro.models.actions import SendMessageAction
from rpft.rapidpro.models.nodes import BasicNode, CallWebhookNode


class TestNodes(unittest.TestCase):
    def test_basic_node(self):
        self.basic_node = BasicNode()
        self.basic_node.add_action(
            SendMessageAction(
                text="test_message_1", quick_replies=["qr01", "qr02", "qr03"]
            )
        )
        self.basic_node.add_action(
            SendMessageAction(
                text="test_image_message", attachments=["image:image url"]
            )
        )
        self.basic_node.add_action(
            SendMessageAction(
                text="test_audio_message", attachments=["audio:audio url"]
            )
        )
        self.basic_node.update_default_exit("test_destination_uuid")

        render_output = self.basic_node.render()

        action_names = [a["text"] for a in render_output["actions"]]

        action_1_index = action_names.index("test_message_1")
        action_2_index = action_names.index("test_image_message")
        action_3_index = action_names.index("test_audio_message")

        self.assertEqual(
            render_output["actions"][action_1_index]["text"], "test_message_1"
        )
        self.assertEqual(
            render_output["actions"][action_2_index]["text"], "test_image_message"
        )
        self.assertEqual(
            render_output["actions"][action_3_index]["text"], "test_audio_message"
        )

    def test_webhook_node(self):
        node = CallWebhookNode(
            body="Webhook Body",
            method="GET",
            url="http://localhost:49998/?cmd=success",
            headers={"Authorization": "Token AAFFZZHH"},
            result_name="webhook_result",
        )
        render_output = node.render()

        action_exp = {
            "type": "call_webhook",
            "body": "Webhook Body",
            "method": "GET",
            "url": "http://localhost:49998/?cmd=success",
            "headers": {"Authorization": "Token AAFFZZHH"},
            "result_name": "webhook_result",
        }

        render_output["actions"][0].pop("uuid")
        self.assertEqual(render_output["actions"][0], action_exp)
        router = render_output["router"]
        self.assertEqual(router["type"], "switch")
        self.assertEqual(router["operand"], "@results.webhook_result.category")
        self.assertEqual(len(router["cases"]), 1)
        self.assertEqual(len(router["categories"]), 2)



import unittest

from rpft.parsers.common.rowparser import RowParser
from rpft.parsers.creation.flowrowmodel import FlowRowModel
from tests.mocks import MockCellParser


input1 = {
    "row_id": "1",
    "type": "send_message",
    "from": "start",
    "condition_value": "3",
    "condition_var": "@fields.name",
    "condition_type": "has_phrase",
    "condition_name": "",
    "message_text": "Text of message",
    "choices": ["Answer 1", "Answer 2"],
    # CellParser is unaware of types, so '' is NOT turned into a list by the cell parser
    "ui_position": "",
}

output1_exp = FlowRowModel(
    **{
        "row_id": "1",
        "type": "send_message",
        "edges": [
            {
                "from_": "start",
                "condition": {
                    "value": "3",
                    "variable": "@fields.name",
                    "type": "has_phrase",
                    "name": "",
                },
            }
        ],
        "mainarg_message_text": "Text of message",
        "choices": ["Answer 1", "Answer 2"],
        "ui_position": [],  # The RowParser converts '' into []
    }
)


input2 = {
    "row_id": "1",
    "type": "send_message",
    "from": "start",
    "condition_value": ["3", "5"],
    "condition_var": "@fields.name",
    "condition_type": "has_phrase",
    "condition_name": "",
    "message_text": "Text of message",
}

output2_exp = FlowRowModel(
    **{
        "row_id": "1",
        "type": "send_message",
        "edges": [
            {
                "from_": "start",
                "condition": {
                    "value": "3",
                    "variable": "@fields.name",
                    "type": "has_phrase",
                    "name": "",
                },
            },
            {
                "from_": "start",
                "condition": {
                    "value": "5",
                    "variable": "@fields.name",
                    "type": "has_phrase",
                    "name": "",
                },
            },
        ],
        "mainarg_message_text": "Text of message",
    }
)


# This is an interesting case.
# It might be that in practice, we would rather that this is interpreted
# as one from with a condition and one unconditional one.
# To realize that, however, we'd have to decouple from from its conditions.
input3 = {
    "row_id": "1",
    "type": "send_message",
    "from": ["start", "5"],
    "condition_value": "3",
    "condition_var": "@fields.name",
    "condition_type": "has_phrase",
    "condition_name": "",
    "message_text": "Text of message",
}

output3_exp = FlowRowModel(
    **{
        "row_id": "1",
        "type": "send_message",
        "edges": [
            {
                "from_": "start",
                "condition": {
                    "value": "3",
                    "variable": "@fields.name",
                    "type": "has_phrase",
                    "name": "",
                },
            },
            {
                "from_": "5",
                "condition": {
                    "value": "3",
                    "variable": "@fields.name",
                    "type": "has_phrase",
                    "name": "",
                },
            },
        ],
        "mainarg_message_text": "Text of message",
    }
)


input4 = {
    "row_id": "1",
    "type": "add_to_group",
    "from": "start",
    "message_text": "Group Name",
}

output4_exp = FlowRowModel(
    **{
        "row_id": "1",
        "type": "add_to_group",
        "edges": [
            {
                "from_": "start",
            },
        ],
        "mainarg_groups": ["Group Name"],
    }
)


input5 = {
    "row_id": "1",
    "type": "go_to",
    "from": "start",
    "message_text": ["5", "2", "3"],
}

output5_exp = FlowRowModel(
    **{
        "row_id": "1",
        "type": "go_to",
        "edges": [
            {
                "from_": "start",
            },
        ],
        "mainarg_destination_row_ids": ["5", "2", "3"],
    }
)


input6 = {
    "row_id": "1",
    "type": "send_message",
    "edges": [["start", ["5"]], ["start", ["3", "@fields.name", "has_phrase"]]],
    "message_text": "Text of message",
}

output6_exp = FlowRowModel(
    **{
        "row_id": "1",
        "type": "send_message",
        "edges": [
            {
                "from_": "start",
                "condition": {"value": "5"},
            },
            {
                "from_": "start",
                "condition": {
                    "value": "3",
                    "variable": "@fields.name",
                    "type": "has_phrase",
                    "name": "",
                },
            },
        ],
        "mainarg_message_text": "Text of message",
    }
)


class TestDifferentWays(unittest.TestCase):
    def setUp(self):
        self.parser = RowParser(FlowRowModel, MockCellParser())

    def test_input_1(self):
        output1 = self.parser.parse_row(input1)
        self.assertEqual(output1, output1_exp)

    def test_input_2(self):
        output2 = self.parser.parse_row(input2)
        self.assertEqual(output2, output2_exp)

    def test_input_3(self):
        output3 = self.parser.parse_row(input3)
        self.assertEqual(output3, output3_exp)

    def test_input_4(self):
        output4 = self.parser.parse_row(input4)
        self.assertEqual(output4, output4_exp)

    def test_input_5(self):
        output5 = self.parser.parse_row(input5)
        self.assertEqual(output5, output5_exp)

    def test_input_6(self):
        output6 = self.parser.parse_row(input6)
        self.assertEqual(output6, output6_exp)



import csv
import tablib

from tests import TESTS_ROOT


def get_dict_from_csv(csv_file_path):
    with open(TESTS_ROOT / csv_file_path) as csv_file:
        csv_reader = csv.DictReader(csv_file)
        return [row for row in csv_reader]


def get_table_from_file(file_path, file_format="csv"):
    # format: Import file format.
    #     Supported file formats as supported by tablib,
    #     see https://tablib.readthedocs.io/en/stable/formats.html
    # Returns:
    #     tablib.Dataset
    with open(TESTS_ROOT / file_path) as data_stream:
        return tablib.import_set(data_stream, format=file_format)


def csv_join(*args):
    return "\n".join(args) + "\n"



import logging
import re

from jinja2 import ChainableUndefined, Environment, contextfilter
from jinja2.nativetypes import NativeEnvironment


LOGGER = logging.getLogger(__name__)


class CellParserError(Exception):
    pass


class CellParser:

    # Separators by level
    # Note: split_into_lists currently assumes there are exactly two separators.
    SEPARATORS = ["|", ";"]
    ESCAPE_CHARACTER = "\\"

    def escape_string(string):
        string = string.replace(
            CellParser.ESCAPE_CHARACTER, CellParser.ESCAPE_CHARACTER * 2
        )
        for sep in CellParser.SEPARATORS:
            string = string.replace(sep, CellParser.ESCAPE_CHARACTER + sep)
        return string

    @contextfilter
    def evaluate_string(context, string):
        return eval(string, {}, context)

    def __init__(self):
        self.env = Environment(undefined=ChainableUndefined)
        self.env.filters["escape"] = CellParser.escape_string
        self.env.filters["eval"] = CellParser.evaluate_string
        self.native_env = NativeEnvironment(
            variable_start_string="{@",
            variable_end_string="@}",
            undefined=ChainableUndefined,
        )
        self.native_env.filters["escape"] = CellParser.escape_string
        self.native_env.filters["eval"] = CellParser.evaluate_string

    def split_into_lists(self, string):
        l1 = self.split_by_separator(string, CellParser.SEPARATORS[0])
        if type(l1) is str:
            output = self.split_by_separator(string, CellParser.SEPARATORS[1])
        else:
            output = [self.split_by_separator(s, CellParser.SEPARATORS[1]) for s in l1]
        return unescape(output)

    def split_by_separator(self, string, sep):
        pos = 0
        sep_locations = []
        while pos < len(string):
            c = string[pos]
            if c == CellParser.ESCAPE_CHARACTER:
                pos += 1
            elif c == sep:
                sep_locations.append(pos)
            pos += 1
        if not sep_locations:
            # No separators found: return a string, not a list
            return string
        else:
            if len(string) - 1 in sep_locations:
                # Special case: Last character is a separator.
                # Here we don't put '' at the end of the list
                locations = [-1] + sep_locations[:-1] + [len(string) - 1]
            else:
                locations = [-1] + sep_locations + [len(string)]
            return [
                string[locations[i] + 1 : locations[i + 1]]
                for i in range(len(locations) - 1)
            ]

    def parse(self, value, context={}):
        value, is_object = self.parse_as_string(value, context)

        return value if is_object else self.split_into_lists(value)

    def parse_as_string(self, value, context={}):
        """
        Render template in cell value, optionally, using the given context.

        Returns a tuple consisting of the result of the rendering and a boolean that
        indicates to the caller whether the parsing result represents an object that is
        not to be processed any further. For example, templates that return Python types
        other than string i.e {@ ... @}.
        """
        is_object = False

        if value is None:
            return "", is_object

        stripped = str(value).strip()

        if context is None or (not context and "{" not in stripped):
            return stripped, is_object

        env = self.env

        if stripped.startswith("{@") and stripped.endswith("@}"):
            env = self.native_env
            is_object = True

            # Ensure this is a single template, not e.g. '{@ x @} {@ y @}'
            if "{@" in stripped[2:]:
                raise Exception(
                    'Cell may not contain nested "{{@" templates.'
                    f'Cell content: "{stripped}"'
                )

        try:
            return env.from_string(stripped).render(context), is_object
        except Exception as e:
            raise Exception(
                f'Error while parsing cell "{stripped}" with context "{context}":'
                f" {str(e)}"
            )

    def join_from_lists(self, value, depth=0):
        if type(value) is str:
            return CellParser.escape_string(value)
        elif type(value) in [int, bool, float]:
            return str(value)
        elif type(value) is list:
            if depth > len(CellParser.SEPARATORS):
                raise CellParserError(
                    "Error while converting nested list into string: "
                    "Input list is nested too deeply."
                )
            if len(value) == 1:
                # Trailing separator to distinguish 1-element lists from basic types
                return (
                    self.join_from_lists(value[0], depth=depth + 1)
                    + CellParser.SEPARATORS[depth]
                )
            return CellParser.SEPARATORS[depth].join(
                [self.join_from_lists(e, depth=depth + 1) for e in value]
            )
        else:
            raise CellParserError(
                "Error while converting nested list into string: "
                "Invalid type of list element."
            )


def unescape(nested_list):
    """Unescape escaped characters"""
    return (
        re.sub(r"\\(.{1})", r"\g<1>", nested_list.strip())
        if type(nested_list) is str
        else [unescape(item) for item in nested_list]
    )



import logging

from rpft.rapidpro.models.campaigns import Campaign, CampaignEvent
from rpft.logger.logger import logging_context


LOGGER = logging.getLogger(__name__)


class CampaignParser:
    def __init__(self, name, group_name, rows):
        self.campaign = Campaign(name, group_name=group_name)
        self.rows = rows

    def parse(self):
        for row_idx, row in enumerate(self.rows):
            with logging_context(f"row {row_idx+2}"):
                message = None
                base_language = None
                if row.message:
                    message = {"eng": row.message}
                    base_language = row.base_language or "eng"
                delivery_hour = -1
                if row.delivery_hour:
                    delivery_hour = int(row.delivery_hour)
                try:
                    event = CampaignEvent(
                        int(row.offset),
                        row.unit,
                        row.event_type,
                        delivery_hour,
                        row.start_mode,
                        relative_to_label=row.relative_to,
                        flow_name=row.flow or None,
                        message=message,
                        base_language=base_language,
                    )
                    self.campaign.add_event(event)
                except ValueError as e:
                    raise Exception(str(e))
        return self.campaign



class RapidProActionError(Exception):
    "Raised if some parameter of a RapidProAction is invalid."

    pass


class RapidProRouterError(Exception):
    "Raised if some parameter of a RapidProRouter is invalid."

    pass



from collections import defaultdict
from typing import List, ForwardRef, _eval_type
from pydoc import locate
from pydantic import create_model

from rpft.parsers.common.rowparser import (
    ParserModel,
    RowParser,
    RowParserError,
    get_field_name,
    is_list_type,
    is_parser_model_type,
    str_to_bool,
)


def type_from_string(string):
    if not string:
        # By default, assume str
        return str
    basic_type = locate(string)
    if basic_type:
        return basic_type
    try:
        inferred_type = _eval_type(ForwardRef(string), globals(), globals())
    except NameError as e:
        raise RowParserError(f'Error while parsing type "{string}": {str(e)}')
    return inferred_type


def get_value_for_type(type, value=None):
    if is_list_type(type):
        # We do not support default values for lists.
        return []
    if is_parser_model_type(type):
        # We do not support default values for ParserModel.
        return type()
    if value is not None:
        if type is bool:
            return str_to_bool(value)
        return type(value)
    return type()


def infer_type(string):
    if RowParser.TYPE_ANNOTATION_SEPARATOR not in string:
        return type_from_string("")
    # Take the stuff between colon and equal sign
    prefix, suffix = string.split(RowParser.TYPE_ANNOTATION_SEPARATOR, 1)
    return type_from_string(suffix.split(RowParser.DEFAULT_VALUE_SEPARATOR)[0].strip())


def infer_default_value(type, string):
    if RowParser.DEFAULT_VALUE_SEPARATOR not in string:
        # Return the default value for the given type
        return get_value_for_type(type)
    prefix, suffix = string.split(RowParser.DEFAULT_VALUE_SEPARATOR, 1)
    return get_value_for_type(type, suffix.strip())


def parse_header_annotations(string):
    inferred_type = infer_type(string)
    return inferred_type, infer_default_value(inferred_type, string)


def represents_integer(string):
    try:
        _ = int(string)
        return True
    except ValueError:
        return False


def dict_to_list(dict):
    out = [None] * (max(dict.keys()) + 1)
    for k, v in dict.items():
        out[k] = v
    return out


def model_from_headers(name, headers):
    return model_from_headers_rec(name, headers)[0]


def model_from_headers_rec(name, headers):
    # Returns a model and a default value
    fields = {}
    complex_fields = defaultdict(list)
    for header in headers:
        if RowParser.HEADER_FIELD_SEPARATOR in header:
            field, subheader = header.split(RowParser.HEADER_FIELD_SEPARATOR, 1)
            complex_fields[field].append(subheader)
        else:
            field = get_field_name(header)
            field_type, default_value = parse_header_annotations(header)
            fields[field] = (field_type, default_value)
    for field, subheaders in complex_fields.items():
        # Assign model and default value
        fields[field] = model_from_headers_rec(name.title() + field.title(), subheaders)

    # In case the model that we're creating is a list,
    # all its fields are numbers (indices).
    list_model = None
    list_default_values = {}
    for field, value in fields.items():
        if represents_integer(field):
            # We do not check whether the models for each list entry match.
            # We just take one of them.
            list_model = value[0]
            # Index shift: because in the headers, we count from 1
            list_default_values[int(field) - 1] = value[1]
    if list_model is not None:
        return List[list_model], dict_to_list(list_default_values)

    # If the model we're creating is not a list, it's a class
    model = create_model(name.title(), __base__=ParserModel, **fields)
    return model, get_value_for_type(model)



import json
from unittest import TestCase

from tablib import Dataset

from rpft.converters import to_json
from rpft.parsers.sheets import AbstractSheetReader, Sheet


class TestReaderToJson(TestCase):
    def test_something(self):
        reader = MockSheetReader(
            {
                "sheet1": Sheet(
                    reader=None,
                    name="sheet1",
                    table=Dataset(
                        ("row1_col1", "row1_col2", "row1_col3"),
                        headers=["col1", "col2", "col3"],
                    ),
                ),
                "sheet2": Sheet(
                    reader=None,
                    name="sheet2",
                    table=Dataset(
                        ("row1_col1", "row1_col2", "row1_col3"),
                        headers=["col1", "col2", "col3"],
                    ),
                ),
            }
        )
        self.assertDictEqual(
            json.loads(to_json(reader)),
            {
                "meta": {
                    "version": "0.1.0",
                },
                "sheets": {
                    "sheet1": [
                        {
                            "col1": "row1_col1",
                            "col2": "row1_col2",
                            "col3": "row1_col3",
                        },
                    ],
                    "sheet2": [
                        {
                            "col1": "row1_col1",
                            "col2": "row1_col2",
                            "col3": "row1_col3",
                        },
                    ],
                },
            },
        )


class MockSheetReader(AbstractSheetReader):
    def __init__(self, sheets):
        self._sheets = sheets



import networkx as nx
import tablib


class RowDataSheet:
    def __init__(self, row_parser, rows, target_headers=set(), excluded_headers=set()):
        """
        Class to export list of models to spreadsheet.

        Note: If the model supports remapping where multiple fields of different
        types are mapped to the same column header, that column header needs to
        be specified in target_headers.

        Args:
            rows (list[ParserModel]): a list of RowModel instances
            row_parser (RowParser): parser to use for converting rows to flat dicts.
            target_headers (set[str]): Complex type fields (ParserModels, lists, dicts)
                whose content should be represented in the output as single columns.
                A trailing asterisk may be used to specify multiple fields at once,
                such as `list.*` and `field.*`.
            excluded_headers (set[str]): Fields to exclude from the output. Same format
                as target_headers.
        """
        self.row_parser = row_parser
        self.rows = rows
        self.target_headers = target_headers
        self.excluded_headers = excluded_headers

    def export(self, filename, file_format="csv"):
        """
        Export a list of RowModel instances to file.

        Args:
            filename: destination filename
            format: Export file format.
                Supported file formats as supported by tablib,
                see https://tablib.readthedocs.io/en/stable/formats.html
        """
        data = self.convert_to_tablib()
        exported_data = data.export(file_format)
        if type(exported_data) is str:
            # There is a strange bug where on Windows, csv files would get
            # exported with \r\r\n linebreaks. The below fixes this.
            exported_data = exported_data.replace("\r", "")
            exported_data = exported_data.encode("utf-8")
        with open(filename, "wb") as f:
            f.write(exported_data)

    def convert_to_tablib(self):
        """
        Convert a list of RowModel instances to tablib.Dataset.

        Return:
          A tablib.Dataset representation of the data.
        """

        data = tablib.Dataset()
        data.headers = self._get_headers()
        for row in self.rows:
            row_dict = self.row_parser.unparse_row(
                row, self.target_headers, self.excluded_headers
            )
            data.append([row_dict.get(header, "") for header in data.headers])
        return data

    def _get_headers(self):
        """
        Get an ordered list of column headers.
        Each row contains a subset of the final set of column headers.
        These subsets need to be merged while respecting the relative order
        within each row. Note: The resulting set of headers is unique,
        however, their order is not guaranteed to be unique.
        TODO: A better approach would be to use the DataModel of the rows
        to uniquely infer the order of the headers.
        Return:
            A list of strings representing the column headers of the sheet.
        """

        # Create a graph (representing a poset) whose nodes are the column headers,
        # and whose edges A -> B represent that column header A should come before
        # column header B.
        header_graph = nx.DiGraph()
        for row in self.rows:
            row_dict = self.row_parser.unparse_row(
                row, self.target_headers, self.excluded_headers
            )
            k_prev = None
            # For each pair of consecutive headers in this row, add an edge.
            for k, _ in row_dict.items():
                if k_prev:
                    header_graph.add_edge(k_prev, k)
                k_prev = k

        # We now get a linear order of our headers from this poset graph
        # by doing a topological sort.
        try:
            ordering = list(nx.topological_sort(header_graph))
        except nx.exception.NetworkXUnfeasible:
            raise ValueError("Inconsistent ordering of headers in provided rows.")
        return ordering



import unittest
import json

from rpft.rapidpro.models.actions import Action, Group
from rpft.rapidpro.models.common import Exit
from rpft.rapidpro.models.routers import RouterCase, RouterCategory, BaseRouter
from rpft.rapidpro.models.nodes import BaseNode
from rpft.rapidpro.models.containers import FlowContainer, RapidProContainer
from rpft.rapidpro.models.campaigns import Campaign, CampaignEvent
from tests import TESTS_ROOT


class TestImportExport(unittest.TestCase):
    def setUp(self) -> None:
        self.data_dir = TESTS_ROOT / "data"

    def test_all_action_types(self):
        actionFilenamesList = self.data_dir.glob("actions/*.json")
        for filename in actionFilenamesList:
            with open(filename, "r") as f:
                action_data = json.load(f)
            action = Action.from_dict(action_data)
            render_output = action.render()
            self.assertEqual(render_output, action_data, msg=filename)

    def test_exits(self):
        with open(self.data_dir / "exits/exit.json", "r") as f:
            data = json.load(f)
        exit = Exit.from_dict(data)
        render_output = exit.render()
        self.assertEqual(render_output, data)

    def test_groups(self):
        with open(self.data_dir / "groups/group.json", "r") as f:
            data = json.load(f)
        group = Group.from_dict(data)
        render_output = group.render()
        self.assertEqual(render_output, data)

    def test_cases(self):
        with open(self.data_dir / "routers/case.json", "r") as f:
            data = json.load(f)
        case = RouterCase.from_dict(data)
        render_output = case.render()
        self.assertEqual(render_output, data)

    def test_categories(self):
        with open(self.data_dir / "routers/category.json", "r") as f:
            data = json.load(f)
        with open(self.data_dir / "routers/exits.json", "r") as f:
            exit_data = json.load(f)
        exits = [Exit.from_dict(exit) for exit in exit_data]
        category = RouterCategory.from_dict(data, exits)
        render_output = category.render()
        self.assertEqual(render_output, data)

    def test_all_router_types(self):
        # Note: These test cases assume that the default category
        # within switch routers is always the last one.
        # This might change once we support "Expired" categories
        self.maxDiff = None
        routerFilenamesList = self.data_dir.glob("routers/router_*.json")
        for filename in routerFilenamesList:
            with open(filename, "r") as f:
                router_data = json.load(f)
            with open(self.data_dir / "routers/exits.json", "r") as f:
                exit_data = json.load(f)
            exits = [Exit.from_dict(exit) for exit in exit_data]
            router = BaseRouter.from_dict(router_data, exits)
            render_output = router.render()
            self.assertEqual(render_output, router_data, msg=filename)

    def test_all_node_types(self):
        self.maxDiff = None
        nodeFilenamesList = self.data_dir.glob("nodes/node_*.json")
        for filename in nodeFilenamesList:
            with open(filename, "r") as f:
                node_data = json.load(f)
            node = BaseNode.from_dict(node_data)
            render_output = node.render()
            self.assertEqual(render_output, node_data, msg=filename)

    def test_flow_containers(self):
        self.maxDiff = None
        containerFilenamesList = self.data_dir.glob("containers/flow_container_*.json")
        for filename in containerFilenamesList:
            with open(filename, "r") as f:
                container_data = json.load(f)
            container = FlowContainer.from_dict(container_data)
            render_output = container.render()
            # TODO: compare nodes/UI element-wise, for smaller error output?
            self.assertEqual(render_output, container_data, msg=filename)

    def test_rapidpro_containers(self):
        self.maxDiff = None
        containerFilenamesList = self.data_dir.glob(
            "containers/rapidpro_container_*.json"
        )
        for filename in containerFilenamesList:
            with open(filename, "r") as f:
                container_data = json.load(f)
            container = RapidProContainer.from_dict(container_data)
            render_output = container.render()
            self.assertEqual(render_output, container_data, msg=filename)

    def test_rapidproprev_container_triggers(self):
        # Previous versions of RapidPro had a different trigger formats.
        # Check compatibility in this test.
        self.maxDiff = None
        with open(
            self.data_dir / "containers/rapidpro_container_trigger.json", "r"
        ) as f:
            expected_container_data = json.load(f)
        with open(
            self.data_dir / "containers/rapidproprev_container_trigger.json", "r"
        ) as f:
            container_data = json.load(f)
        container = RapidProContainer.from_dict(container_data)
        render_output = container.render()
        self.assertEqual(render_output, expected_container_data)

    def test_campaign_events(self):
        self.maxDiff = None

        for filename in self.data_dir.glob("campaigns/event_*.json"):
            with open(filename, "r") as f:
                data = json.load(f)

            self.assertEqual(
                CampaignEvent.from_dict(data).render(),
                data,
                msg=filename,
            )

    def test_campaigns(self):
        self.maxDiff = None

        for filename in self.data_dir.glob("campaigns/campaign_*.json"):
            with open(filename, "r") as f:
                data = json.load(f)

            self.assertEqual(
                Campaign.from_dict(data).render(),
                data,
                msg=filename,
            )



from typing import List

from rpft.parsers.creation.datarowmodel import DataRowModel
from rpft.parsers.common.rowparser import ParserModel


class CustomModel(ParserModel):
    # Because this does not directly define the content of a datasheet,
    # in inherits from ParserModel and not DataRowModel.
    happy: str = ""
    sad: str = ""


class NestedRowModel(DataRowModel):
    # Because this defines the content of a datasheet,
    # it inherits DataRowModel, which gives it the ID column.
    value1: str = ""
    custom_field: CustomModel = CustomModel()  # Default value is an empty custom model


class ListRowModel(DataRowModel):
    # Because this defines the content of a datasheet,
    # it inherits DataRowModel, which gives it the ID column.
    list_value: List[str] = []



import logging

from rpft.logger.logger import logging_context
from rpft.rapidpro.models.triggers import Trigger


LOGGER = logging.getLogger(__name__)


class TriggerParser:
    def __init__(self, sheet_name, rows):
        self.sheet_name = sheet_name
        self.rows = rows

    def parse(self):
        triggers = []
        for row_idx, row in enumerate(self.rows):
            with logging_context(f"row {row_idx+2}"):
                try:
                    trigger = Trigger(
                        row.type,
                        row.keywords,
                        row.channel,
                        row.match_type,
                        flow_name=row.flow,
                        group_names=row.groups,
                        group_uuids=[],
                        exclude_group_names=row.exclude_groups,
                        exclude_group_uuids=[],
                    )
                    triggers.append(trigger)
                except ValueError as e:
                    raise Exception(str(e))
        return triggers



import copy
import json
import tablib
from unittest import TestCase

from rpft.parsers.common.cellparser import CellParser
from rpft.parsers.common.rowparser import RowParser
from rpft.parsers.creation.flowrowmodel import FlowRowModel
from rpft.parsers.creation.flowparser import FlowParser
from rpft.rapidpro.models.actions import Group, AddContactGroupAction
from rpft.rapidpro.models.containers import RapidProContainer, FlowContainer
from rpft.rapidpro.models.nodes import BasicNode
from rpft.rapidpro.simulation import (
    Context,
    find_destination_uuid,
    find_node_by_uuid,
    traverse_flow,
)

from tests import TESTS_ROOT
from tests.mocks import MockSheetParser
from tests.row_data import (
    get_conditional_node_from_1,
    get_message_with_templating,
    get_unconditional_node_from_1,
    get_start_row,
)
from tests.utils import get_dict_from_csv, get_table_from_file


class TestParsing(TestCase):
    def setUp(self) -> None:
        self.row_parser = RowParser(FlowRowModel, CellParser())

    def get_render_output_from_file(self, flow_name, filename):
        self.rows = [
            self.row_parser.parse_row(row) for row in get_dict_from_csv(filename)
        ]
        return self.get_render_output(flow_name, self.rows)

    def get_render_output(self, flow_name, input_rows):
        return (
            FlowParser(
                RapidProContainer(),
                flow_name=flow_name,
                sheet_parser=MockSheetParser(input_rows),
            )
            .parse()
            .render()
        )

    def test_send_message(self):
        output = self.get_render_output("send_message", [get_start_row()])

        action = output["nodes"][0]["actions"][0]
        self.assertEqual(action["type"], "send_msg")
        self.assertEqual(action["text"], "Text of message")
        self.assertEqual(len(action["attachments"]), 0)
        self.assertEqual(action["quick_replies"], ["Answer 1", "Answer 2"])

    def test_send_message_with_template(self):
        output = self.get_render_output(
            "send_message_with_template",
            [get_start_row(), get_message_with_templating()],
        )

        self.assertNotIn("templating", output["nodes"][0]["actions"][0])

        action = output["nodes"][1]["actions"][0]
        self.assertIn("uuid", action["templating"])
        self.assertEqual(action["templating"]["template"]["name"], "template name")
        self.assertEqual(action["templating"]["template"]["uuid"], "template uuid")
        self.assertEqual(action["templating"]["variables"], ["var1", "var2"])

    def test_linear(self):
        row = get_unconditional_node_from_1()
        row.edges[0].from_ = ""  # implicit continue from last node

        output = self.get_render_output("linear", [get_start_row(), row])

        node_0, node_1 = output["nodes"]
        self.assertEqual(node_1["actions"][0]["text"], "Unconditional message")
        self.assertEqual(len(node_0["exits"]), 1)
        self.assertIsNone(node_0.get("router"))
        self.assertEqual(node_0["exits"][0]["destination_uuid"], node_1["uuid"])
        self.assertEqual(node_1["exits"][0]["destination_uuid"], None)

    def test_only_conditional(self):
        output = self.get_render_output(
            "only_conditional",
            [get_start_row(), get_conditional_node_from_1()],
        )

        self.assertEqual(len(output["nodes"]), 3)
        node_0, node_1, node_2 = output["nodes"]
        self.assertEqual(node_2["actions"][0]["text"], "Message if @fields.name == 3")
        self.assertEqual(len(node_1["exits"]), 2)
        self.assertIsNone(node_0.get("router"))
        self.assertIsNotNone(node_1.get("router"))
        self.assertEqual(node_0["exits"][0]["destination_uuid"], node_1["uuid"])
        self.assertEqual(node_1["exits"][0]["destination_uuid"], node_2["uuid"])
        self.assertEqual(node_1["exits"][1]["destination_uuid"], None)
        self.assertEqual(node_2["exits"][0]["destination_uuid"], None)

    def test_split1(self):
        output = self.get_render_output(
            "split",
            [
                get_start_row(),
                get_unconditional_node_from_1(),
                get_conditional_node_from_1(),
            ],
        )

        self.assert_split(output)

    def test_split2(self):
        output = self.get_render_output(
            "split",
            [
                get_start_row(),
                get_conditional_node_from_1(),
                get_unconditional_node_from_1(),
            ],
        )

        self.assert_split(output)

    def assert_split(self, output):
        node_start = output["nodes"][0]
        node_switch = find_node_by_uuid(
            output, node_start["exits"][0]["destination_uuid"]
        )
        default_destination = find_destination_uuid(
            node_switch, Context(variables={"@fields.name": "5"})
        )
        node_2 = find_node_by_uuid(output, default_destination)
        cond_destination = find_destination_uuid(
            node_switch, Context(variables={"@fields.name": "3"})
        )
        node_3 = find_node_by_uuid(output, cond_destination)

        self.assertEqual(node_3["actions"][0]["text"], "Message if @fields.name == 3")
        self.assertEqual(node_2["actions"][0]["text"], "Unconditional message")
        self.assertEqual(len(node_switch["exits"]), 2)
        self.assertIsNone(node_start.get("router"))
        self.assertIsNotNone(node_switch.get("router"))

    def test_no_switch_node_rows(self):
        output = self.get_render_output_from_file(
            "no_switch_node", "input/no_switch_nodes.csv"
        )

        # Check that node UUIDs are maintained
        nodes = output["nodes"]
        all_node_uuids = [row.node_uuid for row in self.rows]
        # Rows 0,1,2,3 and rows -3,-2 are actions joined into a single node.
        expected_node_uuids = all_node_uuids[3:-3] + all_node_uuids[-2:]
        self.assertEqual(
            expected_node_uuids,
            [node["uuid"] for node in nodes],
        )

        self.assertEqual(output["name"], "no_switch_node")
        self.assertEqual(output["type"], "messaging")
        self.assertEqual(output["language"], "eng")

        self.assertEqual(len(nodes), 8)

        node_0 = nodes[0]
        self.assertEqual(len(node_0["actions"]), 4)
        self.assertEqual(node_0["actions"][0]["type"], "send_msg")
        self.assertEqual(node_0["actions"][0]["text"], "this is a send message node")
        self.assertEqual(len(node_0["actions"][0]["attachments"]), 0)
        self.assertIn("qr1", node_0["actions"][0]["quick_replies"])
        self.assertIn("qr2", node_0["actions"][0]["quick_replies"])

        self.assertEqual(node_0["actions"][1]["type"], "send_msg")
        self.assertEqual(node_0["actions"][1]["text"], "message with image")
        self.assertEqual(len(node_0["actions"][1]["attachments"]), 1)
        self.assertEqual(node_0["actions"][1]["attachments"][0], "image:image u")
        self.assertEqual(len(node_0["actions"][1]["quick_replies"]), 0)

        self.assertEqual(node_0["actions"][2]["type"], "send_msg")
        self.assertEqual(node_0["actions"][2]["text"], "message with audio")
        self.assertEqual(len(node_0["actions"][2]["attachments"]), 1)
        self.assertEqual(node_0["actions"][2]["attachments"][0], "audio:audio u")
        self.assertEqual(len(node_0["actions"][2]["quick_replies"]), 0)

        self.assertEqual(node_0["actions"][3]["type"], "send_msg")
        self.assertEqual(node_0["actions"][3]["text"], "message with video")
        self.assertEqual(len(node_0["actions"][3]["attachments"]), 1)
        self.assertEqual(node_0["actions"][3]["attachments"][0], "video:video u")
        self.assertEqual(len(node_0["actions"][3]["quick_replies"]), 0)

        node_1 = nodes[1]
        self.assertEqual(node_0["exits"][0]["destination_uuid"], node_1["uuid"])
        self.assertEqual(len(node_1["actions"]), 1)
        self.assertEqual(node_1["actions"][0]["type"], "set_contact_field")
        self.assertEqual(node_1["actions"][0]["field"]["key"], "test_variable")
        self.assertEqual(node_1["actions"][0]["field"]["name"], "test variable")
        self.assertEqual(node_1["actions"][0]["value"], "test value")

        node_2 = nodes[2]
        self.assertEqual(node_1["exits"][0]["destination_uuid"], node_2["uuid"])
        self.assertEqual(len(node_2["actions"]), 1)
        self.assertEqual(node_2["actions"][0]["type"], "add_contact_groups")
        self.assertEqual(len(node_2["actions"][0]["groups"]), 1)
        self.assertEqual(node_2["actions"][0]["groups"][0]["name"], "test group")

        node_3 = nodes[3]
        self.assertEqual(node_2["exits"][0]["destination_uuid"], node_3["uuid"])
        self.assertEqual(len(node_3["actions"]), 1)
        self.assertEqual("remove_contact_groups", node_3["actions"][0]["type"])
        self.assertEqual(len(node_3["actions"][0]["groups"]), 1)
        self.assertEqual("test group", node_3["actions"][0]["groups"][0]["name"])

        # Make sure it's the same group
        self.assertEqual(
            node_2["actions"][0]["groups"][0]["uuid"],
            node_3["actions"][0]["groups"][0]["uuid"],
        )

        node_4 = nodes[4]
        self.assertEqual(node_3["exits"][0]["destination_uuid"], node_4["uuid"])
        self.assertEqual(len(node_4["actions"]), 1)
        self.assertEqual("set_run_result", node_4["actions"][0]["type"])
        self.assertEqual("result name", node_4["actions"][0]["name"])
        self.assertEqual("result value", node_4["actions"][0]["value"])
        self.assertEqual("my_result_cat", node_4["actions"][0]["category"])

        node_5 = nodes[5]
        self.assertEqual(node_4["exits"][0]["destination_uuid"], node_5["uuid"])
        self.assertEqual(len(node_5["actions"]), 2)
        self.assertEqual(node_5["actions"][0]["type"], "set_contact_language")
        self.assertEqual(node_5["actions"][0]["language"], "eng")
        self.assertEqual(node_5["actions"][1]["type"], "set_contact_name")
        self.assertEqual(node_5["actions"][1]["name"], "John Doe")

        node_6 = nodes[6]
        self.assertEqual(node_5["exits"][0]["destination_uuid"], node_6["uuid"])

        node_7 = nodes[7]
        self.assertEqual(len(node_7["actions"]), 1)
        self.assertEqual("remove_contact_groups", node_7["actions"][0]["type"])
        self.assertEqual(len(node_7["actions"][0]["groups"]), 0)
        self.assertTrue(node_7["actions"][0]["all_groups"])
        self.assertIsNone(node_7["exits"][0]["destination_uuid"])

        # Check UI positions/types of the first two nodes
        render_ui = output["_ui"]["nodes"]
        self.assertIn(node_0["uuid"], render_ui)
        pos0 = render_ui[node_0["uuid"]]["position"]
        self.assertEqual((280, 73), (pos0["left"], pos0["top"]))
        self.assertEqual("execute_actions", render_ui[node_0["uuid"]]["type"])
        self.assertIn(node_1["uuid"], render_ui)
        pos1 = render_ui[node_1["uuid"]]["position"]
        self.assertEqual((280, 600), (pos1["left"], pos1["top"]))
        self.assertEqual("execute_actions", render_ui[node_1["uuid"]]["type"])

    def test_switch_node_rows(self):
        output = self.get_render_output_from_file(
            "switch_node", "input/switch_nodes.csv"
        )

        # Check that node UUIDs are maintained
        nodes = output["nodes"]
        self.assertEqual(
            [row.node_uuid for row in self.rows],
            [node["uuid"] for node in nodes],
        )

        # Check that No Response category is created even if not connected
        last_wait_node = nodes[12]
        self.assertEqual(
            "No Response", last_wait_node["router"]["categories"][-1]["name"]
        )

        # TODO: Ideally, there should be more explicit tests here.
        # At least the functionality is covered by the integration tests simulating the
        # flow.

        render_ui = output["_ui"]["nodes"]

        def f_uuid(i):
            return nodes[i]["uuid"]

        def f_uipos_dict(i):
            return render_ui[f_uuid(i)]["position"]

        def f_uipos(i):
            return (f_uipos_dict(i)["left"], f_uipos_dict(i)["top"])

        def f_uitype(i):
            return render_ui[f_uuid(i)]["type"]

        self.assertIn(f_uuid(0), render_ui)
        self.assertEqual((340, 0), f_uipos(0))
        self.assertEqual((360, 180), f_uipos(1))
        self.assertEqual((840, 1200), f_uipos(12))
        self.assertEqual((740, 300), f_uipos(13))
        self.assertEqual("wait_for_response", f_uitype(0))
        self.assertEqual("split_by_subflow", f_uitype(1))
        self.assertEqual("split_by_expression", f_uitype(2))
        self.assertEqual("split_by_contact_field", f_uitype(3))
        self.assertEqual("split_by_run_result", f_uitype(4))
        self.assertEqual("split_by_groups", f_uitype(5))
        self.assertEqual("wait_for_response", f_uitype(6))
        self.assertEqual("split_by_random", f_uitype(7))
        self.assertEqual("execute_actions", f_uitype(8))
        self.assertEqual("execute_actions", f_uitype(9))
        self.assertEqual("wait_for_response", f_uitype(12))

        # Ensure that wait_for_response cases are working as intended
        node6 = nodes[6]
        categories = node6["router"]["categories"]
        self.assertEqual(len(categories), 3)
        self.assertEqual(categories[0]["name"], "A")
        self.assertEqual(categories[1]["name"], "Other")
        self.assertEqual(categories[2]["name"], "No Response")
        self.assertEqual(len(node6["router"]["cases"]), 1)

    def test_groups_and_flows(self):
        # We check that references flows and group are assigned uuids consistently
        tiny_uuid = "00000000-acec-434f-bc7c-14c584fc4bc8"
        test_uuid = "8224bfe2-acec-434f-bc7c-14c584fc4bc8"
        other_uuid = "12345678-acec-434f-bc7c-14c584fc4bc8"
        test_group_dict = {"name": "test group", "uuid": test_uuid}
        other_group_dict = {"name": "other group", "uuid": other_uuid}
        tiny_flow_dict = {"name": "tiny_flow", "uuid": tiny_uuid}

        # Make a flow with a single group node (no UUIDs), and put it into new container
        node = BasicNode()
        node.add_action(
            AddContactGroupAction([Group("test group"), Group("other group")])
        )
        tiny_flow = FlowContainer("tiny_flow", uuid=tiny_uuid)
        tiny_flow.add_node(node)
        container = RapidProContainer(groups=[Group("other group", other_uuid)])
        container.add_flow(tiny_flow)

        # Add flow from sheet into container
        rows = [
            self.row_parser.parse_row(row)
            for row in get_dict_from_csv("input/groups_and_flows.csv")
        ]
        FlowParser(
            container,
            flow_name="groups_and_flows",
            sheet_parser=MockSheetParser(rows),
        ).parse()

        # Render also invokes filling in all the flow/group UUIDs
        output = container.render()

        # Ensure container groups are complete and have correct UUIDs
        self.assertIn(test_group_dict, output["groups"])
        self.assertIn(other_group_dict, output["groups"])
        # These UUIDs are inferred from the sheet/the container groups, respectively
        self.assertEqual(
            output["flows"][0]["nodes"][0]["actions"][0]["groups"],
            [test_group_dict, other_group_dict],
        )

        nodes = output["flows"][1]["nodes"]
        # This UUID appears only in a later occurrence of the group in the sheet
        self.assertEqual(nodes[0]["actions"][0]["groups"], [test_group_dict])
        # This UUID is missing from the sheet, but explicit in the flow definition
        self.assertEqual(nodes[1]["actions"][0]["flow"], tiny_flow_dict)
        # This UUID is explicit in the sheet
        self.assertEqual(nodes[2]["actions"][0]["groups"], [test_group_dict])
        # This UUID appears in a previous occurrence of the group in the sheet
        self.assertEqual(
            nodes[3]["router"]["cases"][0]["arguments"],
            [test_uuid, "test group"],
        )
        # This UUID was part of the groups in the container, but not in the sheet
        self.assertEqual(nodes[6]["actions"][0]["groups"], [other_group_dict])

        tiny_flow.uuid = "something else"
        with self.assertRaises(ValueError):
            # The enter_flow node has a different uuid than the flow
            container.validate()

        tiny_flow.uuid = tiny_uuid
        container.flows[1].nodes[2].actions[0].groups[0].uuid = "something else"
        with self.assertRaises(ValueError):
            # The group is referenced by 2 different UUIDs
            container.validate()


class TestBlocks(TestCase):

    def setUp(self) -> None:
        self.row_parser = RowParser(FlowRowModel, CellParser())

    def render_output(self, table, context=None):
        return (
            FlowParser(
                RapidProContainer(),
                "basic loop",
                (
                    table
                    if isinstance(table, tablib.Dataset)
                    else tablib.import_set(table, format="csv")
                ),
                context=context or {},
            )
            .parse()
            .render()
        )

    def assert_messages(self, output, expected, context=None):
        self.assertEqual(
            traverse_flow(output, context or Context()),
            list(zip(["send_msg"] * len(expected), expected)),
        )

    def assert_actions(self, output, expected, context=None):
        self.assertEqual(
            traverse_flow(output, context or Context()),
            expected,
        )


class TestWebhook(TestBlocks):

    def test_basic_webhook(self):
        table_data = (
            "row_id,type,from,message_text,webhook.url,webhook.method,webhook.headers,save_name\n"  # noqa: E501
            ",call_webhook,start,Webhook Body,http://localhost:49998/?cmd=success,GET,Authorization;Token AAFFZZHH|,webhook_result\n"  # noqa: E501
        )
        action_exp = {
            "type": "call_webhook",
            "body": "Webhook Body",
            "method": "GET",
            "url": "http://localhost:49998/?cmd=success",
            "headers": {"Authorization": "Token AAFFZZHH"},
            "result_name": "webhook_result",
        }

        output = self.render_output(table_data)

        node = output["nodes"][0]
        node["actions"][0].pop("uuid")
        self.assertEqual(node["actions"][0], action_exp)

    def test_webhook_default_args(self):
        table_data = (
            "row_id,type,from,message_text,webhook.url,webhook.method,webhook.headers,save_name\n"  # noqa: E501
            ",call_webhook,start,Webhook Body,http://localhost:49998/?cmd=success,,,webhook_result\n"  # noqa: E501
        )

        output = self.render_output(table_data)

        node = output["nodes"][0]
        self.assertEqual(node["actions"][0]["headers"], {})
        self.assertEqual(node["actions"][0]["method"], "POST")

    def test_webhook_connectivity(self):
        table = (
            "row_id,type,from,condition,message_text,webhook.url,webhook.method,webhook.headers,save_name\n"  # noqa: E501
            "0,call_webhook,start,,Webhook Body,URL,,,webhook_result\n"
            ",send_message,0,Success,Webhook Success,,,,\n"
            ",send_message,0,,Webhook Failure,,,,\n"
        )

        node = self.render_output(table)["nodes"][0]

        self.assertEqual(node["actions"][0]["headers"], {})
        self.assertEqual(node["actions"][0]["method"], "POST")

        # Also check that the connectivity is correct
        self.assert_actions(
            self.render_output(table),
            [
                ("call_webhook", "URL"),
                ("send_msg", "Webhook Success"),
            ],
            context=Context(inputs=["Success"]),
        )
        self.assert_actions(
            self.render_output(table),
            [
                ("call_webhook", "URL"),
                ("send_msg", "Webhook Failure"),
            ],
            context=Context(inputs=["Failure"]),
        )


class TestLoops(TestBlocks):

    def test_basic_loop(self):
        table = (
            "row_id,type,from,loop_variable,message_text\n"
            "1,begin_for,start,i,1;2;3\n"
            ",send_message,,,{{i}}. Some text\n"
            ",end_for,,,\n"
        )

        nodes = self.render_output(table)["nodes"]

        self.assertEqual(len(nodes), 3)
        self.assertEqual(nodes[0]["actions"][0]["type"], "send_msg")
        self.assertEqual(nodes[0]["actions"][0]["text"], "1. Some text")
        self.assertEqual(nodes[1]["actions"][0]["text"], "2. Some text")
        self.assertEqual(nodes[2]["actions"][0]["text"], "3. Some text")

        # Also check that the connectivity is correct
        self.assert_messages(
            self.render_output(table),
            ["1. Some text", "2. Some text", "3. Some text"],
        )

    def test_enumerate(self):
        table_data = (
            "row_id,type,from,loop_variable,message_text\n"
            "1,begin_for,start,text;i,A;B;C\n"
            ",send_message,,,{{i+1}}. {{text}}\n"
            ",end_for,,,\n"
        )
        render_output = self.render_output(table_data)
        nodes = render_output["nodes"]
        self.assertEqual(len(nodes), 3)
        self.assertEqual(nodes[0]["actions"][0]["type"], "send_msg")
        self.assertEqual(nodes[0]["actions"][0]["text"], "1. A")
        self.assertEqual(nodes[1]["actions"][0]["text"], "2. B")
        self.assertEqual(nodes[2]["actions"][0]["text"], "3. C")

    def test_skip_loop_when_list_is_empty(self):
        table = tablib.Dataset(
            ("send_message", "", "", "Start"),
            ("begin_for", "", "i", "{@ [] @}"),
            ("send_message", "", "", "{{ i }}"),
            ("end_for", "", "", ""),
            ("send_message", "", "", "End"),
            headers=("type", "from", "loop_variable", "message_text"),
        )
        self.assert_messages(
            self.render_output(table),
            ["Start", "End"],
        )

    def test_one_element_loop(self):
        table_data = (
            "row_id,type,from,loop_variable,message_text\n"
            "1,begin_for,start,i,label\n"
            ",send_message,,,{{i}}. Some text\n"
            ",end_for,,,\n"
        )
        render_output = self.render_output(table_data)
        nodes = render_output["nodes"]
        self.assertEqual(len(nodes), 1)
        self.assertEqual(nodes[0]["actions"][0]["type"], "send_msg")
        self.assertEqual(nodes[0]["actions"][0]["text"], "label. Some text")

    def test_nested_loop(self):
        table = (
            "row_id,type,from,loop_variable,message_text\n"
            "1,begin_for,start,i,1;2\n"
            ",begin_for,,j,A;B\n"
            ",send_message,,,{{i}}{{j}}. Some text\n"
            ",end_for,,,\n"
            ",end_for,,,\n"
        )
        expected = [
            "1A. Some text",
            "1B. Some text",
            "2A. Some text",
            "2B. Some text",
        ]

        output = self.render_output(table)

        self.assert_messages(output, expected)

    def test_loop_within_other_nodes(self):
        table = (
            "row_id,type,from,loop_variable,message_text\n"
            "1,send_message,start,,Starting text\n"
            "2,begin_for,1,i,1;2\n"
            ",send_message,,,{{i}}. Some text\n"
            ",end_for,,,\n"
            ",send_message,,,Following text\n"
        )
        expected = [
            "Starting text",
            "1. Some text",
            "2. Some text",
            "Following text",
        ]

        output = self.render_output(table)

        self.assert_messages(output, expected)

    def test_nested_loop_with_other_nodes(self):
        table = (
            "row_id,type,from,loop_variable,message_text\n"
            "1,begin_for,start,i,1;2\n"
            ",begin_for,,j,A;B\n"
            ",send_message,,,{{i}}{{j}}. Some text\n"
            ",end_for,,,\n"
            ",send_message,,,End of inner loop\n"
            ",end_for,,,\n"
            ",send_message,,,End of outer loop\n"
        )
        expected = [
            "1A. Some text",
            "1B. Some text",
            "End of inner loop",
            "2A. Some text",
            "2B. Some text",
            "End of inner loop",
            "End of outer loop",
        ]

        output = self.render_output(table)

        self.assert_messages(output, expected)

    def test_loop_with_explicit_following_node(self):
        table = (
            "row_id,type,from,loop_variable,message_text\n"
            "2,begin_for,,i,1;2\n"
            ",send_message,,,{{i}}. Some text\n"
            ",end_for,,,\n"
            ",send_message,2,,Following text\n"
        )
        expected = ["1. Some text", "2. Some text", "Following text"]

        output = self.render_output(table)

        self.assert_messages(output, expected)

    def test_loop_with_goto(self):
        table = (
            "row_id,type,from,condition,loop_variable,message_text\n"
            "2,begin_for,start,,i,1;2\n"
            ",send_message,,,,{{i}}. Text\n"
            ",end_for,,,,\n"
            "3,wait_for_response,2,,,\n"
            ",go_to,3,hello,,2\n"
            ",send_message,3,,,Following text\n"
        )

        self.assert_messages(
            self.render_output(table),
            ["1. Text", "2. Text", "Following text"],
            Context(inputs=["goodbye"]),
        )

        self.assert_messages(
            self.render_output(table),
            ["1. Text", "2. Text", "1. Text", "2. Text", "Following text"],
            Context(inputs=["hello", "goodbye"]),
        )

    def test_loop_with_goto_into_middle_of_loop(self):
        table = (
            "row_id,type,from,condition,loop_variable,message_text\n"
            "2,begin_for,start,,i,1;2\n"
            "item{{i}},send_message,,,,{{i}}. Text\n"
            ",end_for,,,,\n"
            "3,wait_for_response,2,,,\n"
            ",go_to,3,hello,,item2\n"
            ",send_message,3,,,Following text\n"
        )

        self.assert_messages(
            self.render_output(table),
            ["1. Text", "2. Text", "Following text"],
            Context(inputs=["goodbye"]),
        )
        self.assert_messages(
            self.render_output(table),
            ["1. Text", "2. Text", "2. Text", "Following text"],
            Context(inputs=["hello", "goodbye"]),
        )

    def test_loop_over_object(self):
        class TestObj:
            def __init__(self, value):
                self.value = value

        context = {"test_objs": [TestObj("1"), TestObj("2"), TestObj("A")]}
        table = (
            "row_id,type,from,loop_variable,message_text\n"
            "2,begin_for,start,obj,{@test_objs@}\n"
            ",send_message,,,Value: {{obj.value}}\n"
            ",end_for,,,\n"
        )

        self.assert_messages(
            self.render_output(table, context),
            ["Value: 1", "Value: 2", "Value: A"],
        )

    def test_loop_over_range(self):
        table = (
            "row_id,type,from,loop_variable,message_text\n"
            "2,begin_for,,i,{@range(5)@}\n"
            ",send_message,,,{{i}}. Text\n"
            ",end_for,,,\n"
        )

        self.assert_messages(
            self.render_output(table),
            ["0. Text", "1. Text", "2. Text", "3. Text", "4. Text"],
        )


class TestConditionals(TestBlocks):
    def test_block_within_other_nodes(self):
        table = (
            "row_id,type,from,message_text\n"
            ",send_message,start,Starting text\n"
            ",begin_block,,\n"
            ",send_message,,Some text\n"
            ",end_block,,\n"
            ",send_message,,Following text\n"
        )

        self.assert_messages(
            self.render_output(table),
            ["Starting text", "Some text", "Following text"],
        )

    def test_block_with_explicit_from(self):
        table = (
            "row_id,type,from,message_text\n"
            ",send_message,start,Start\n"
            "X,begin_block,,\n"
            ",send_message,,Text 1\n"
            ",send_message,,Text 2\n"
            ",end_block,,\n"
            ",send_message,X,Following\n"
        )

        self.assert_messages(
            self.render_output(table),
            ["Start", "Text 1", "Text 2", "Following"],
        )

    def test_block_with_goto(self):
        table = (
            "row_id,type,from,condition,message_text\n"
            "2,begin_block,start,,\n"
            ",send_message,,,Some text\n"
            ",end_block,,,\n"
            "3,wait_for_response,2,,\n"
            ",go_to,3,hello,2\n"
            ",send_message,3,,Following text\n"
        )

        self.assert_messages(
            self.render_output(table),
            ["Some text", "Following text"],
            Context(inputs=["goodbye"]),
        )
        self.assert_messages(
            self.render_output(table),
            ["Some text", "Some text", "Following text"],
            Context(inputs=["hello", "goodbye"]),
        )

    def test_basic_if(self):
        table = (
            "row_id,type,from,include_if,message_text\n"
            ",send_message,,,text1\n"
            ",send_message,,FALSE,text2\n"
            ",send_message,,something,text3\n"
            ",send_message,,False,text4\n"
            ",send_message,,{{1 == 0}},text5\n"
            ",send_message,,{@1 == 0@},text6\n"
            ",send_message,,{@1 == 1@},text7\n"
        )

        self.assert_messages(
            self.render_output(table),
            ["text1", "text3", "text7"],
        )

    def test_excluded_block_within_other_nodes(self):
        table = (
            "row_id,type,from,include_if,message_text\n"
            ",send_message,start,,Starting text\n"
            ",begin_block,,FALSE,\n"
            ",send_message,,,Skipped text\n"
            ",send_message,,TRUE,Skipped text 2\n"  # Should be skipped anyway
            ",end_block,,,\n"
            ",send_message,,,Following text\n"
        )

        self.assert_messages(
            self.render_output(table),
            ["Starting text", "Following text"],
        )

    def test_excluded_for_block(self):
        table = (
            "row_id,type,from,include_if,message_text\n"
            ",begin_for,,FALSE,1;2\n"  # No loop var; but it's not parsed anyway
            ",send_message,,,Skipped text\n"
            ",end_for,,,\n"
            ",send_message,,,Following text\n"
        )

        self.assert_messages(
            self.render_output(table),
            ["Following text"],
        )

    def test_excluded_block_with_nested_stuff(self):
        table = (
            "row_id,type,from,include_if,message_text\n"
            ",begin_block,,FALSE,\n"
            ",begin_block,,,\n"
            ",send_message,,,Skipped text\n"
            ",begin_for,,,1;2;3\n"  # No loop var; but it's not parsed anyway
            ",send_message,,,{{i}}. Some text\n"
            ",end_for,,,\n"
            ",end_block,,,\n"
            ",begin_for,,,A;B\n"
            ",send_message,,,{{i}}. Some other text\n"
            ",end_for,,,\n"
            ",end_block,,,\n"
            ",send_message,,,Following text\n"
        )

        self.assert_messages(
            self.render_output(table),
            ["Following text"],
        )


class TestMultiExitBlocks(TestBlocks):

    def test_split_by_value(self):
        table = (
            "row_id,type,from,condition,message_text\n"
            "X,begin_block,,,\n"
            "1,split_by_value,,,@my_field\n"
            ",send_message,1,Value,It has the value\n"
            ",end_block,,,\n"
            ",send_message,X,,Following text\n"
        )

        self.assert_messages(
            self.render_output(table),
            ["It has the value", "Following text"],
            Context(variables={"@my_field": "Value"}),
        )
        self.assert_messages(
            self.render_output(table),
            ["Following text"],
            Context(variables={"@my_field": "Other"}),
        )

    def test_split_by_value_hard_loose_exit(self):
        table = (
            "row_id,type,from,condition,message_text\n"
            "X,begin_block,,,\n"
            "1,split_by_value,,,@my_field\n"
            ",send_message,1,Value,It has the value\n"
            ",hard_exit,1,Value2,\n"
            ",loose_exit,1,Value3,\n"
            ",end_block,,,\n"
            ",send_message,X,,Following text\n"
        )

        self.assert_messages(
            self.render_output(table),
            ["It has the value", "Following text"],
            Context(variables={"@my_field": "Value"}),
        )
        self.assert_messages(
            self.render_output(table),
            [],
            Context(variables={"@my_field": "Value2"}),
        )
        self.assert_messages(
            self.render_output(table),
            ["Following text"],
            Context(variables={"@my_field": "Value3"}),
        )
        self.assert_messages(
            self.render_output(table),
            ["Following text"],
            Context(variables={"@my_field": "Other"}),
        )

    def test_wait_for_response(self):
        table = (
            "row_id,type,from,condition,message_text,no_response\n"
            "X,begin_block,,,,\n"
            "1,wait_for_response,,,,60\n"
            ",send_message,1,Value,It has the value,\n"
            ",send_message,1,No Response,No Response,\n"
            ",end_block,,,\n"
            ",send_message,1,,Other,\n"
            ",send_message,X,,Following text,\n"
        )

        self.assert_messages(
            self.render_output(table),
            ["It has the value", "Following text"],
            Context(inputs=["Value"]),
        )
        self.assert_messages(
            self.render_output(table),
            ["No Response", "Following text"],
            Context(inputs=[None]),
        )
        self.assert_messages(
            self.render_output(table),
            ["Other"],
            Context(inputs=["Something else"]),
        )

    def test_wait_for_response_hard_exits(self):
        table = (
            "row_id,type,from,condition,message_text,no_response\n"
            "X,begin_block,,,,\n"
            "1,wait_for_response,,,,60\n"
            ",send_message,1,Value,It has the value,\n"
            ",hard_exit,,,,\n"
            ",send_message,1,No Response,No Response,\n"
            ",hard_exit,,,,\n"
            ",end_block,,,\n"
            ",send_message,X,,Following text,\n"
        )

        self.assert_messages(
            self.render_output(table),
            ["It has the value"],
            Context(inputs=["Value"]),
        )
        self.assert_messages(
            self.render_output(table),
            ["No Response"],
            Context(inputs=[None]),
        )
        self.assert_messages(
            self.render_output(table),
            ["Following text"],
            Context(inputs=["Something else"]),
        )

    def test_enter_flow(self):
        table = (
            "row_id,type,from,condition,message_text\n"
            ",send_message,start,,Starting text\n"
            "X,begin_block,,,\n"
            "1,start_new_flow,,,Some_flow\n"
            ",send_message,1,completed,Completed\n"
            ",end_block,,,\n"
            ",send_message,X,,Following text\n"
        )

        self.assert_actions(
            self.render_output(table),
            [
                ("send_msg", "Starting text"),
                ("enter_flow", "Some_flow"),
                ("send_msg", "Completed"),
                ("send_msg", "Following text"),
            ],
            Context(inputs=["completed"]),
        )
        self.assert_actions(
            self.render_output(table),
            [
                ("send_msg", "Starting text"),
                ("enter_flow", "Some_flow"),
                ("send_msg", "Following text"),
            ],
            Context(inputs=["expired"]),
        )


class TestNoOpRow(TestBlocks):

    def test_basic_noop(self):
        table = (
            "row_id,type,from,message_text\n"
            ",send_message,,Start message\n"
            ",no_op,,\n"
            ",send_message,,End message\n"
        )

        self.assert_messages(
            self.render_output(table),
            ["Start message", "End message"],
        )

    def test_multientry_noop(self):
        table = (
            "row_id,type,from,condition,message_text\n"
            "1,wait_for_response,,,\n"
            "2,send_message,1,A,Text A\n"
            "3,send_message,1,,Other\n"
            ",no_op,2;3,,\n"
            ",send_message,,,End message\n"
        )

        self.assert_messages(
            self.render_output(table),
            ["Text A", "End message"],
            context=Context(inputs=["A"]),
        )
        self.assert_messages(
            self.render_output(table),
            ["Other", "End message"],
            context=Context(inputs=["something"]),
        )

    def test_multiexit_noop(self):
        table = (
            "row_id,type,from,condition_value,condition_variable,message_text\n"
            ",send_message,,,,Start message\n"
            "1,no_op,,,,\n"
            ",send_message,1,A,@field,Text A\n"
            ",send_message,1,,@field,Other\n"
            ",send_message,1,B,@field,Text B\n"
        )

        self.assert_messages(
            self.render_output(table),
            ["Start message", "Text A"],
            context=Context(variables={"@field": "A"}),
        )
        self.assert_messages(
            self.render_output(table),
            ["Start message", "Text B"],
            context=Context(variables={"@field": "B"}),
        )
        self.assert_messages(
            self.render_output(table),
            ["Start message", "Other"],
            context=Context(variables={"@field": "something"}),
        )

    def test_multiexit_noop2(self):
        # only two rows are swapped, compared to previous case
        table = (
            "row_id,type,from,condition_value,condition_variable,message_text\n"
            ",send_message,,,,Start message\n"
            "1,no_op,,,,\n"
            ",send_message,1,,@field,Other\n"
            ",send_message,1,A,@field,Text A\n"
            ",send_message,1,B,@field,Text B\n"
        )

        self.assert_messages(
            self.render_output(table),
            ["Start message", "Text A"],
            context=Context(variables={"@field": "A"}),
        )
        self.assert_messages(
            self.render_output(table),
            ["Start message", "Text B"],
            context=Context(variables={"@field": "B"}),
        )
        self.assert_messages(
            self.render_output(table),
            ["Start message", "Other"],
            context=Context(variables={"@field": "something"}),
        )

    def test_multientryexit_noop(self):
        # only two rows are swapped, compared to previous case
        table = (
            "row_id,type,from,condition_value,condition_variable,message_text\n"
            "1,wait_for_response,,,,\n"
            "2,send_message,1,A,,Text 1A\n"
            "3,send_message,1,,,Other\n"
            "4,no_op,2;3,,,\n"
            ",send_message,4,A,@field,Text 2A\n"
            ",send_message,4,,@field,Other\n"
            ",send_message,4,B,@field,Text 2B\n"
        )

        self.assert_messages(
            self.render_output(table),
            ["Text 1A", "Text 2A"],
            context=Context(inputs=["A"], variables={"@field": "A"}),
        )
        self.assert_messages(
            self.render_output(table),
            ["Text 1A", "Text 2B"],
            context=Context(inputs=["A"], variables={"@field": "B"}),
        )
        self.assert_messages(
            self.render_output(table),
            ["Text 1A", "Other"],
            context=Context(inputs=["A"], variables={"@field": "something"}),
        )
        self.assert_messages(
            self.render_output(table),
            ["Other", "Text 2A"],
            context=Context(inputs=["something"], variables={"@field": "A"}),
        )
        self.assert_messages(
            self.render_output(table),
            ["Other", "Text 2B"],
            context=Context(inputs=["something"], variables={"@field": "B"}),
        )
        self.assert_messages(
            self.render_output(table),
            ["Other", "Other"],
            context=Context(inputs=["something"], variables={"@field": "something"}),
        )

    def test_noop_in_block_loose(self):
        # only two rows are swapped, compared to previous case
        table = (
            "row_id,type,from,condition_value,condition_variable,message_text\n"
            "X,begin_block,,,,\n"
            ",send_message,,,,Start message\n"
            "1,no_op,,,,\n"
            ",end_block,,,,\n"
            ",send_message,,,,End message\n"
        )

        self.assert_messages(
            self.render_output(table),
            ["Start message", "End message"],
        )

    def test_noop_in_block_notloose(self):
        # only two rows are swapped, compared to previous case
        table = (
            "row_id,type,from,condition_value,condition_variable,message_text\n"
            "X,begin_block,,,,\n"
            ",send_message,,,,Start message\n"
            "1,no_op,,,,\n"
            ",send_message,,,,End message\n"
            ",end_block,,,,\n"
            ",send_message,,,,End message 2\n"
        )

        self.assert_messages(
            self.render_output(table),
            ["Start message", "End message", "End message 2"],
        )

    def test_noop_in_block2(self):
        # only two rows are swapped, compared to previous case
        table = (
            "row_id,type,from,condition_value,condition_variable,message_text\n"
            "X,begin_block,,,,\n"
            ",send_message,,,,Start message\n"
            "1,no_op,,,,\n"
            ",loose_exit,1,,@field,\n"
            ",hard_exit,1,A,@field,\n"
            ",loose_exit,1,B,@field,\n"
            ",end_block,,,,\n"
            ",send_message,,,,End Message\n"
        )

        self.assert_messages(
            self.render_output(table),
            ["Start message"],
            context=Context(variables={"@field": "A"}),
        )
        self.assert_messages(
            self.render_output(table),
            ["Start message", "End Message"],
            context=Context(variables={"@field": "B"}),
        )
        self.assert_messages(
            self.render_output(table),
            ["Start message", "End Message"],
            context=Context(variables={"@field": "something"}),
        )

    def test_multientry_block(self):
        table = (
            "row_id,type,from,condition,message_text\n"
            "1,wait_for_response,start,,\n"
            "2,send_message,1,A,Message A\n"
            "3,send_message,1,,Other\n"
            "X,begin_block,2;3,,\n"
            ",send_message,,,Some text 1\n"
            ",end_block,,,\n"
            ",send_message,,,Following text\n"
        )

        self.assert_messages(
            self.render_output(table),
            ["Message A", "Some text 1", "Following text"],
            context=Context(inputs=["A"]),
        )
        self.assert_messages(
            self.render_output(table),
            ["Other", "Some text 1", "Following text"],
            context=Context(inputs=["something"]),
        )


class TestFlowParser(TestCase):

    def assert_flow(self, filename, flow_name, context):
        # Generate a flow from sheet
        output_1 = (
            FlowParser(
                RapidProContainer(),
                flow_name,
                get_table_from_file(filename),
            )
            .parse()
            .render()
        )

        # Load the expected output flow
        with open(TESTS_ROOT / "output/all_test_flows.json", "r") as file:
            expected_flow = next(
                iter(
                    flow
                    for flow in json.load(file)["flows"]
                    if flow["name"] == flow_name
                )
            )

        # Ensure the generated flow and expected flow are functionally equivalent
        expected_actions = traverse_flow(expected_flow, copy.deepcopy(context))
        self.assertEqual(
            traverse_flow(output_1, copy.deepcopy(context)),
            expected_actions,
        )

        # Convert the expected output into a flow and then into a sheet
        new_rows = FlowContainer.from_dict(expected_flow).to_rows()

        # Now convert the sheet back into a flow
        output_2 = (
            FlowParser(
                RapidProContainer(),
                flow_name=flow_name,
                sheet_parser=MockSheetParser(new_rows),
            )
            .parse()
            .render()
        )

        # Ensure the new generated flow and expected flow are functionally equivalent
        self.assertEqual(
            traverse_flow(output_2, copy.deepcopy(context)),
            expected_actions,
        )

    def test_no_switch_nodes(self):
        self.assert_flow(
            "input/no_switch_nodes.csv",
            "no_switch_nodes",
            Context(),
        )

    def test_no_switch_nodes_without_row_ids(self):
        self.assert_flow(
            "input/no_switch_nodes_without_row_ids.csv",
            "no_switch_nodes",
            Context(),
        )

    def test_switch_nodes(self):
        self.assert_flow(
            "input/switch_nodes.csv",
            "switch_nodes",
            Context(inputs=["b", "expired", "Success"]),
        )
        self.assert_flow(
            "input/switch_nodes.csv",
            "switch_nodes",
            Context(inputs=["b", "expired", "Failure", "Success"]),
        )
        self.assert_flow(
            "input/switch_nodes.csv",
            "switch_nodes",
            Context(inputs=["a", "completed"], variables={"expression": "not a"}),
        )
        self.assert_flow(
            "input/switch_nodes.csv",
            "switch_nodes",
            Context(
                inputs=["a", "completed"],
                group_names=["wrong group"],
                variables={
                    "expression": "a",
                    "@contact.name": "a",
                    "@results.result_wfr": "a",
                },
            ),
        )
        self.assert_flow(
            "input/switch_nodes.csv",
            "switch_nodes",
            Context(
                inputs=["a", "completed", "other"],
                group_names=["test group"],
                variables={
                    "expression": "a",
                    "@contact.name": "a",
                    "@results.result_wfr": "a",
                },
            ),
        )
        self.assert_flow(
            "input/switch_nodes.csv",
            "switch_nodes",
            Context(
                inputs=["a", "completed", "a"],
                random_choices=[0],
                group_names=["test group"],
                variables={
                    "expression": "a",
                    "@contact.name": "a",
                    "@results.result_wfr": "a",
                },
            ),
        )
        self.assert_flow(
            "input/switch_nodes.csv",
            "switch_nodes",
            Context(
                inputs=["a", "completed", "a"],
                random_choices=[2],
                group_names=["test group"],
                variables={
                    "expression": "a",
                    "@contact.name": "a",
                    "@results.result_wfr": "a",
                },
            ),
        )
        self.assert_flow(
            "input/switch_nodes.csv",
            "switch_nodes",
            Context(
                inputs=["a", "completed", None, None],
                group_names=["test group"],
                variables={
                    "expression": "a",
                    "@contact.name": "a",
                    "@results.result_wfr": "a",
                },
            ),
        )

    def test_loop_from_start(self):
        self.assert_flow(
            "input/loop_from_start.csv",
            "loop_from_start",
            Context(inputs=["b"]),
        )

        self.assert_flow(
            "input/loop_from_start.csv",
            "loop_from_start",
            Context(inputs=["a", "b"]),
        )

    def test_rejoin(self):
        self.assert_flow(
            "input/rejoin.csv",
            "rejoin",
            Context(random_choices=[0]),
        )

        self.assert_flow(
            "input/rejoin.csv",
            "rejoin",
            Context(random_choices=[1]),
        )

        self.assert_flow(
            "input/rejoin.csv",
            "rejoin",
            Context(random_choices=[2]),
        )

    def test_loop_and_multiple_conditions(self):
        self.assert_flow(
            "input/loop_and_multiple_conditions.csv",
            "loop_and_multiple_conditions",
            Context(inputs=["adfgyh"], random_choices=[0]),
        )

        self.assert_flow(
            "input/loop_and_multiple_conditions.csv",
            "loop_and_multiple_conditions",
            Context(inputs=["other"], random_choices=[0]),
        )

        self.assert_flow(
            "input/loop_and_multiple_conditions.csv",
            "loop_and_multiple_conditions",
            Context(random_choices=[1]),
        )

        self.assert_flow(
            "input/loop_and_multiple_conditions.csv",
            "loop_and_multiple_conditions",
            Context(inputs=["other"], random_choices=[2]),
        )

        self.assert_flow(
            "input/loop_and_multiple_conditions.csv",
            "loop_and_multiple_conditions",
            Context(inputs=["b", "a"], random_choices=[2]),
        )

        self.assert_flow(
            "input/loop_and_multiple_conditions.csv",
            "loop_and_multiple_conditions",
            Context(inputs=["b", "b", "c"], random_choices=[2]),
        )



import unittest

from rpft.parsers.common.rowparser import RowParser
from tests.mocks import MockCellParser
from tests.models import FromWrong


output_instance = {
    "row_id": "5",
    "conditions": [
        {"value": "1", "var": "", "type": "has_phrase", "name": "A"},
        {"value": "2", "var": "", "type": "has_phrase", "name": "B"},
        {"value": "3", "var": "", "type": "has_phrase", "name": ""},
    ],
}

input1 = {
    "row_id": "5",
    "conditions.*.value": ["1", "2", "3"],
    "conditions.*.var": "",
    "conditions.*.type": ["has_phrase", "has_phrase", "has_phrase"],
    "conditions.*.name": ["A", "B", ""],
}

input2 = {
    "row_id": "5",
    "conditions.*.value": ["1", "2", "3"],
    "conditions.*.var": "",
    "conditions.*.type": ["has_phrase", "has_phrase", "has_phrase"],
    "conditions.*.name": ["A", "B"],
}

input3 = {
    "row_id": "5",
    "conditions.*.value": ["1", "2", "3"],
    "conditions.*.var": "",
    "conditions.*.type": "has_phrase",
    "conditions.*.name": ["A", "B", ""],
}

input4 = {
    "row_id": "5",
    "conditions.1": ["1", "", "has_phrase", "A"],
    "conditions.2": ["2", "", "has_phrase", "B"],
    "conditions.3": ["3", "", "has_phrase", ""],
}

input5 = {
    "row_id": "5",
    "conditions.1": ["1", "", "has_phrase", "A"],
    "conditions.2": ["2", "", "has_phrase", "B"],
    "conditions.3": ["3", "", "has_phrase"],
}

input6 = {
    "row_id": "5",
    "conditions.1": [["value", "1"], ["type", "has_phrase"], ["name", "A"]],
    "conditions.2": [["value", "2"], ["type", "has_phrase"], ["name", "B"]],
    "conditions.3": [["value", "3"], ["type", "has_phrase"]],
}

input7 = {
    "row_id": "5",
    "conditions.1": ["1", "", "has_phrase", ["name", "A"]],
    "conditions.2": ["2", "", ["type", "has_phrase"], ["name", "B"]],
    "conditions.3": ["3", ["type", "has_phrase"]],
}

input8 = {
    "row_id": "5",
    "conditions": [
        ["1", "", "has_phrase", "A"],
        ["2", "", "has_phrase", "B"],
        ["3", "", "has_phrase", ""],
    ],
}

input9 = {
    "row_id": "5",
    "conditions.1.value": "1",
    "conditions.1.type": "has_phrase",
    "conditions.1.name": "A",
    "conditions.2.value": "2",
    "conditions.2.type": "has_phrase",
    "conditions.2.name": "B",
    "conditions.3.value": "3",
    "conditions.3.type": "has_phrase",
}


input_single_kwarg = {
    "row_id": "5",
    "conditions.1": ["value", "3"],
}

output_single_kwarg_exp = FromWrong(
    **{
        "row_id": "5",
        "conditions": [
            {"value": "3", "var": "", "type": "", "name": ""},
        ],
    }
)


class TestDifferentWays(unittest.TestCase):
    def setUp(self):
        self.parser = RowParser(FromWrong, MockCellParser())

    def test_different_ways(self):
        inputs = [
            input1,
            input2,
            input3,
            input4,
            input5,
            input6,
            input7,
            input8,
            input9,
        ]
        outputs = []

        for inp in inputs:
            out = self.parser.parse_row(inp)  # We get an instance of the model
            outputs.append(out)

        for out in outputs:
            self.assertEqual(out.model_dump(), output_instance)

    def test_single_kwarg(self):
        output_single_kwarg = self.parser.parse_row(input_single_kwarg)
        self.assertEqual(output_single_kwarg, output_single_kwarg_exp)



import json
import logging
import re
from collections import defaultdict
from functools import singledispatch
from pathlib import Path
from typing import Any

from benedict import benedict
from tablib import Dataset

from rpft.parsers.sheets import AbstractSheetReader, Sheet

LOGGER = logging.getLogger(__name__)

DELIMS = "|;"
PROP_ACCESSOR = "."
META_KEY = "_idems"
TABULATE_KEY = "tabulate"
HEADERS_KEY = "headers"
Table = list[list[str]]
Book = list[tuple[str, Table]]


def bookify(data: dict) -> Book:
    """
    Convert a dict into a 'book' - a list of named tables.
    """
    meta = data.get(META_KEY, {}).get(TABULATE_KEY, {})

    return [(k, tabulate(v, meta.get(k, {}))) for k, v in data.items() if k != META_KEY]


def tabulate(data, meta: dict = {}) -> Table:
    """
    Convert a nested data structure to a tabular form
    """
    headers = meta.get(HEADERS_KEY, []) or list(
        {k: None for item in data for k, _ in item.items()}.keys()
    )
    paths = keypaths(headers)
    rows = []

    for item in data:
        obj = benedict(item)
        rows += [[stringify(obj[kp]) for kp in paths]]

    return [headers] + rows


@singledispatch
def stringify(value, delimiters=DELIMS, **_) -> str:
    s = str(value)

    return s if is_template(s) else re.sub(rf"([{delimiters}])", r"\\\1", s)


@stringify.register
def _(value: dict, delimiters=DELIMS, depth=0) -> str:
    if len(delimiters[depth:]) > 1:
        d1, d2 = delimiters[depth : depth + 2]
    else:
        raise ValueError("Too few delimiters to stringify dict")

    s = f" {d1} ".join(
        f"{stringify(k)}{d2} {stringify(v, delimiters=delimiters, depth=depth + 2)}"
        for k, v in value.items()
    )

    if len(value) == 1:
        s += " " + d1

    return s


@stringify.register
def _(value: list, delimiters=DELIMS, depth=0) -> str:
    d = delimiters[depth] if depth < len(delimiters) else None

    if not d:
        raise ValueError("Too few delimiters to stringify list")

    s = f" {d} ".join(
        stringify(item, delimiters=delimiters, depth=depth + 1) for item in value
    )

    if len(value) == 1:
        s += f" {d}"
    elif value[-1] == "":
        s += d

    return s


@stringify.register
def _(value: tuple, delimiters=DELIMS, depth=0) -> str:
    return stringify(list(value), delimiters=delimiters, depth=depth)


@stringify.register
def _(value: bool, **_) -> str:
    return str(value).lower()


def parse_tables(reader: AbstractSheetReader) -> dict:
    """
    Parse a workbook into a nested structure
    """
    obj = benedict()

    for title, sheet in reader.sheets.items():
        obj.merge(parse_table(title, sheet.table.headers, sheet.table[:]))

    return obj


def parse_table(title: str = None, headers=tuple(), rows=tuple()):
    """
    Parse data in tabular form into a nested structure
    """
    title = title or "table"

    if not headers or not rows:
        return {title: []}

    return create_obj(stream(title, headers, rows))


def stream(title: str = None, headers=tuple(), rows=tuple()):
    yield [META_KEY, TABULATE_KEY, title, HEADERS_KEY], headers

    for i, row in enumerate(rows):
        for h, v in zip(keypaths(headers), row):
            yield [title, i] + h, parse_cell(v)


def keypaths(headers):
    counters = defaultdict(int)
    indexed = []

    for key in headers:
        indexed += [(key, counters[key])]
        counters[key] += 1

    return [keypath(h, i, counters[h]) for h, i in indexed]


def keypath(header, index, count):
    expanded = [normalise_key(k) for k in header.split(PROP_ACCESSOR)]
    i = index if index < count else count - 1

    return expanded + [i] if count > 1 else expanded


def normalise_key(key):
    try:
        return int(key) - 1
    except ValueError:
        return key


def create_obj(pairs):
    obj = benedict()

    for kp, v in pairs:
        obj[kp] = v

    return obj


def parse_cell(s: str, delimiters=DELIMS, depth=0) -> Any:
    if type(s) is not str:
        raise TypeError("Value to convert is not a string")

    clean = s.strip() if s else ""

    try:
        return int(clean)
    except Exception:
        pass

    try:
        return float(clean)
    except Exception:
        pass

    if clean in ("true", "false"):
        return clean == "true"

    if is_template(clean):
        return clean

    d = delimiters[depth] if depth < len(delimiters) else ""
    pattern = rf"(?<!\\)\{d}"

    if d and re.search(pattern, clean):
        seq = [
            parse_cell(item, delimiters=delimiters, depth=depth + 1)
            for item in re.split(pattern, clean)
        ]

        return seq[:-1] if re.search(rf"(?<!\\)\{d}$", clean) else seq

    delims = delimiters[depth + 1 :]

    if delims and re.search(rf"(?<!\\)[{''.join(delims)}]", clean):
        return parse_cell(clean, delimiters=delimiters, depth=depth + 1)

    return re.sub(rf"\\([{delimiters}])", r"\g<1>", clean)


def is_template(s: str) -> bool:
    return bool(re.search(r"{{.*?}}|{@.*?@}|{%.*?%}|@\(.*?\)", s))


class UniJSONReader(AbstractSheetReader):
    def __init__(self, path):
        self.name = path
        self._sheets = {}

        with open(path, "r") as f:
            data = json.load(f)

        for name, content in data.items():
            if name == "_idems":
                continue

            meta = data.get("_idems", {}).get("tabulate", {}).get(name, {})
            table = tabulate(content, meta)
            self._sheets[name] = Sheet(
                reader=self,
                name=name,
                table=Dataset(*table[1:], headers=table[0], title=name),
            )

    @classmethod
    def can_process(cls, location):
        if Path(location).suffix.lower() == ".json":
            with open(location, "r") as f:
                return "content_index" in json.load(f)

        return False



from unittest import TestCase

from rpft.parsers.sheets import CSVSheetReader, Sheet, XLSXSheetReader, JSONSheetReader
from tests import TESTS_ROOT


class Base:
    class SheetReaderTestCase(TestCase):
        def test_reader_named_correctly(self):
            self.assertEqual(self.reader.name, self.expected_reader_name)

        def test_access_existing_sheet_by_name(self):
            sheet: Sheet = self.reader.get_sheet("my_basic_flow")
            self.assertIsInstance(sheet, Sheet)
            self.assertEqual(sheet.reader, self.reader)
            self.assertEqual(sheet.name, "my_basic_flow")
            self.assertEqual(
                sheet.table[0],
                ("", "send_message", "start", "Some text"),
            )

        def test_access_missing_sheet(self):
            self.assertIsNone(self.reader.get_sheet("missing"))


class TestCsvSheetReader(Base.SheetReaderTestCase):
    def setUp(self):
        path = str(TESTS_ROOT / "input/example1/csv_workbook")
        self.reader = CSVSheetReader(path=path)
        self.expected_reader_name = path


class TestXlsxSheetReader(Base.SheetReaderTestCase):
    def setUp(self):
        filename = str(TESTS_ROOT / "input/example1/content_index.xlsx")
        self.reader = XLSXSheetReader(filename=filename)
        self.expected_reader_name = filename


class TestJsonSheetReader(Base.SheetReaderTestCase):
    def setUp(self):
        filename = str(TESTS_ROOT / "input/example1/content_index.json")
        self.reader = JSONSheetReader(filename=filename)
        self.expected_reader_name = filename



import collections


class TagMatcher:
    def __init__(self, params=[]):
        self.include_patterns = collections.defaultdict(list)
        self.exclude_patterns = collections.defaultdict(list)
        if not params:
            return
        current_index = None
        for param in params:
            try:
                val = int(param) - 1
            except ValueError:
                val = None
            if val is not None:
                current_index = val
            else:
                if current_index is None:
                    raise ValueError(
                        "Tags parameter must start with a number indicating the tag"
                        " position"
                    )
                if param[0] == "!":
                    self.exclude_patterns[current_index].append(param[1:])
                else:
                    self.include_patterns[current_index].append(param)
                    # Empty string is accepted by default
                    self.include_patterns[current_index].append("")

    def matches(self, tags):
        matches = True
        for i, tag in enumerate(tags):
            if i in self.include_patterns and tag not in self.include_patterns[i]:
                matches = False
            if i in self.exclude_patterns and tag in self.exclude_patterns[i]:
                matches = False
        return matches



from typing import List
import unittest
from pydantic import create_model, BaseModel

from rpft.parsers.common.model_inference import (
    get_value_for_type,
    infer_default_value,
    infer_type,
    model_from_headers,
    parse_header_annotations,
    type_from_string,
)


class TestModelInference(unittest.TestCase):
    def test_type_from_string(self):
        self.assertEqual(type_from_string(""), str)
        self.assertEqual(type_from_string("str"), str)
        self.assertEqual(type_from_string("int"), int)
        self.assertEqual(type_from_string("list"), list)
        self.assertEqual(type_from_string("List"), List)
        self.assertEqual(type_from_string("List[str]"), List[str])
        self.assertEqual(type_from_string("List[int]"), List[int])
        self.assertEqual(type_from_string("List[List[int]]"), List[List[int]])

    def test_get_value_for_type(self):
        self.assertEqual(get_value_for_type(int), 0)
        self.assertEqual(get_value_for_type(str), "")
        self.assertEqual(get_value_for_type(bool), False)
        self.assertEqual(get_value_for_type(List[str]), [])
        self.assertEqual(get_value_for_type(list), [])

        self.assertEqual(get_value_for_type(int, "5"), 5)
        self.assertEqual(get_value_for_type(str, "abc"), "abc")
        self.assertEqual(get_value_for_type(bool, "True"), True)
        self.assertEqual(get_value_for_type(bool, "TRUE"), True)

    def test_infer_type(self):
        self.assertEqual(infer_type("field:int"), int)
        self.assertEqual(infer_type("field:list"), list)
        self.assertEqual(infer_type("field:List[int]"), List[int])
        self.assertEqual(infer_type("field : list"), list)
        self.assertEqual(infer_type("field:int=5"), int)
        self.assertEqual(infer_type("field:int = 5"), int)
        self.assertEqual(infer_type("field : int = 5"), int)
        self.assertEqual(infer_type("field=5"), str)

    def test_infer_default_value(self):
        self.assertEqual(infer_default_value(int, "field:int"), 0)
        self.assertEqual(infer_default_value(list, "field:list"), [])
        self.assertEqual(infer_default_value(List[int], "field:List[int]"), [])
        self.assertEqual(infer_default_value(list, "field : list"), [])
        self.assertEqual(infer_default_value(int, "field:int=5"), 5)
        self.assertEqual(infer_default_value(int, "field:int = 5"), 5)
        self.assertEqual(infer_default_value(int, "field : int = 5"), 5)
        self.assertEqual(infer_default_value(str, "field=5"), "5")
        self.assertEqual(infer_default_value(str, "field = 5"), "5")

    def test_parse_header_annotations(self):
        self.assertEqual(parse_header_annotations("field:int=5"), (int, 5))

    def compare_models(self, model1, model2, **kwargs):
        self.assertEqual(model1(**kwargs).model_dump(), model2(**kwargs).model_dump())

    def test_model_from_headers(self):
        self.compare_models(
            model_from_headers("mymodel", ["field1"]),
            create_model(
                "Mymodel",
                field1=(str, ""),
            ),
        )
        self.compare_models(
            model_from_headers("mymodel", ["field1:int=5"]),
            create_model(
                "Mymodel",
                field1=(int, 5),
            ),
        )
        self.compare_models(
            model_from_headers("mymodel", ["field1:list"]),
            create_model(
                "Mymodel",
                field1=(list, []),
            ),
        )
        self.compare_models(
            model_from_headers("mymodel", ["field1:list"]),
            create_model(
                "Mymodel",
                field1=(list, []),
            ),
            field1=[1, 2, 3, 4],
        )

        class MySubmodel(BaseModel):
            sub1: str = ""
            sub2: int = 5

        self.compare_models(
            model_from_headers("mymodel", ["field1.sub1", "field1.sub2:int=5"]),
            create_model(
                "Mymodel",
                field1=(MySubmodel, MySubmodel()),
            ),
        )
        self.compare_models(
            model_from_headers("mymodel", ["field1.1", "field1.2"]),
            create_model(
                "Mymodel",
                field1=(list, ["", ""]),
            ),
        )
        self.compare_models(
            model_from_headers("mymodel", ["field1.1:int", "field1.2:int=5"]),
            create_model(
                "Mymodel",
                field1=(list, [0, 5]),
            ),
        )
        self.compare_models(
            model_from_headers(
                "mymodel",
                [
                    "field1.1.1",
                    "field1.1.2=a",
                    "field1.2.1=b",
                    "field1.2.2=c",
                ],
            ),
            create_model(
                "Mymodel",
                field1=(List[List[str]], [["", "a"], ["b", "c"]]),
            ),
        )
        self.compare_models(
            model_from_headers(
                "mymodel",
                [
                    "field1.1.sub1",
                    "field1.1.sub2:int=5",
                    "field1.2.sub1",
                    "field1.2.sub2:int=5",
                ],
            ),
            create_model(
                "Mymodel",
                field1=(List[MySubmodel], [MySubmodel(), MySubmodel()]),
            ),
        )
